#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include "httpd.h"
#include "conf.h"


FileTransferHandler::FileTransferHandler(HTable conf)
{
    location = std::string(table_get_strs(conf, "location", "", NULL));
    directory = std::string(table_get_strs(conf, "directory", "", NULL));
    buildIndexes(table_get_table(conf, "indexes", NULL, NULL));
}

FileTransferHandler::~FileTransferHandler()
{
}


typedef struct
{
    long int size;
    time_t mtime;
} FileInfo;


static bool isDirectory(const std::string& path)
{
    struct stat st;
    
    if (stat(path.c_str(), &st))
        return false;

    if (S_ISDIR(st.st_mode))
        return true;
    else
        return false;
}


static bool isFileExists(const std::string& filename)
{
    std::ifstream ifs(filename.c_str());
    int res = ifs.good();
    ifs.close();
    return res ? true: false;
}


static std::string findIndexFile(const std::string& dir, 
        std::vector<std::string>& indexes)
{
    for (std::vector<std::string>::iterator i = indexes.begin(); i != indexes.end(); 
            i++)
    {
        std::string p = dir + "/" + *i;
        if (isFileExists(p))
            return p;
    }
    return "";
}


static void getFileInfo(const std::string& path, FileInfo& info)
{
    struct stat st;
    
    if (stat(path.c_str(), &st))
        throw HttpException(404, "File disappeared");
    info.size = st.st_size;
    info.mtime = st.st_mtime;
}


static std::string getExtension(const std::string& path)
{
    int len = path.length();
    int i = len - 1;
    while (i >= 0)
    {
        if (path[i] == '.')
            return path.substr(i + 1);
        i--;
    }
    return "";
}


#define BLOCK_SIZE 1024


static void sendFile(HttpdConnection& conn, const std::string& filename, 
        long int size)
{
    int sz = 0;
    FILE *f = fopen(filename.c_str(), "r");
    if (! f)
        throw HttpException(404, "File not found");
    unsigned char buf[BLOCK_SIZE];
    int readed;
    do
    {
        readed = fread(buf, 1, BLOCK_SIZE, f);
        if (readed > 0)
            conn.sendData(buf, readed);
        sz += readed;
    }
    while (readed == BLOCK_SIZE);
    fclose(f);
    if (sz != size)
        throw MasException("File size changed");
}


int FileTransferHandler::handleRequest(HttpdConnection& connection, 
        const ServerURL& url, enum RequestType requestType)
{
    std::string path = url.getPath();
    while ((path.length() > 0) && (path[0] == '/'))
        path = path.substr(1);

    int idx = path.find('\\');
    if ((idx >= 0) && (idx < (int)path.length()))
        throw HttpException(400, "Invalid characters in URL");
    idx = path.find("..");
    if ((idx >= 0) || (idx < (int)path.length()))
        throw HttpException(400, "Invalid characters in URL");
    
    path = directory + "/" + path;
    if (isDirectory(path))
        path = findIndexFile(path, indexes);
    
    if (! isFileExists(path))
        throw HttpException(404, "File not found");
   
    FileInfo info;
    getFileInfo(path, info);
    
    connection.setStatus(200);
    connection.setContentType(connection.getHttpd()->getMimeType(getExtension(path)));
    connection.setContentLength(info.size);
    connection.transferHeaders();
    if (requestType != REQ_HEAD)
        sendFile(connection, path, info.size);
    return 0;
}


void FileTransferHandler::buildIndexes(HTable conf)
{
    if (! conf) return;
    
    HTableIterator iter = table_get_iter(conf);
    while (iter && table_iter_next(iter))
    {
        std::string filename(table_iter_get_strs(iter, NULL));
        if (filename != "")
            indexes.push_back(filename);
    }
    table_free_iter(iter);
}

