#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <windows.h>
#endif
#include <fcntl.h>
#include <algorithm>
#include <fstream>
#include <iostream>
#include "httpd.h"
#include "exceptions.h"
#include "log.h"
#include "utils.h"
#include "server.h"



HttpException::HttpException(int httpNumber, const std::string& msg):
                MasException(msg)
{
    number = httpNumber;
}

        
int HttpException::getHttpCode() const
{
    return number;
}


Httpd::Httpd()
{
    HTable conf = Server::getServer()->getConfig();
    conf = table_get_table(conf, "httpd", NULL, NULL);
    if (! conf)
        throw MasException("Missing httpd server configuration");

    maxConnections = table_get_int(conf, "max_clients", 10000, NULL);
    createAllowed(table_get_table(conf, "allowed_hosts", NULL, NULL));
    createHandlers(table_get_table(conf, "handlers", NULL, NULL));
    std::string mimeFile = table_get_strs(conf, "mime_types", "", NULL);
    mimeExtensions[""] = "application/data";
    if (mimeFile != "")
        loadMimeTypes(mimeFile);
    createSockets(table_get_table(conf, "listen", NULL, NULL));
}


Httpd::~Httpd()
{
    closeSockets();
}


void Httpd::addHandler(const std::string& url, HttpdHandler *handler)
{
    Locker locker(handlersLock);
    handlers.insert(HandlerRecord(url, handler));
}


void Httpd::removeHandler(const std::string& url)
{
    Locker locker(handlersLock);
    handlers.erase(HandlerRecord(url, NULL));
}


static int createSocket(const std::string& addr, int port)
{
    struct sockaddr_in serv_addr;
    int one, sock;

    memset(&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    if (addr == "*")
        serv_addr.sin_addr.s_addr = INADDR_ANY;
    else
        serv_addr.sin_addr.s_addr = Subnet::atoip(addr);
    serv_addr.sin_port = htons((u_short)port);

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        throw MasException("Error creating socket");
#ifdef WIN32
    unsigned long int block = 1;
    ioctlsocket(sock, FIONBIO, &block);
#else
    fcntl(sock, F_SETFL, O_NONBLOCK);
#endif

    one = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*)&one, sizeof(one));
    if (bind(sock, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) == -1)
    {
        shutdown(sock, 2);
        close(sock);
        char buf[30];
        sprintf(buf, "%i", port);
        throw MasException("Error binding socket '" + addr + "', port " + buf);
    }

    if (listen(sock, 20) < 0)
    {
        shutdown(sock, 2);
        close(sock);
        char buf[30];
        sprintf(buf, "%i", port);
        throw MasException("Error listening socket '" + addr + "', port " + buf);
    }

    return sock;
}


void Httpd::createSockets(HTable conf)
{
    if (! conf) return;
    HTableIterator iter = table_get_iter(conf);    
    try
    {
        while (iter && table_iter_next(iter))
        {
            HTable addrTable = table_iter_get_table(iter, NULL);
            std::string addr = std::string(table_get_strs(addrTable, "address", "*", NULL));
            int port = table_get_int(addrTable, "port", 8000, NULL);
            int sock = createSocket(addr, port);
            sockets.push_back(sock);
        }
    }
    catch (MasException& e)
    {
        table_free_iter(iter);
        closeSockets();
        throw e;
    }
    catch (...)
    {
        table_free_iter(iter);
        closeSockets();
        throw MasException("Error creating sockets");
    }
    table_free_iter(iter);
    if (sockets.size() < 1)
        throw MasException("No sockets to listen");
}


void Httpd::createAllowed(HTable conf)
{
    if (! conf) return;
    HTableIterator iter = table_get_iter(conf);
    try
    {
        while (iter && table_iter_next(iter))
        {
            HTable addrTable = table_iter_get_table(iter, NULL);
            std::string addr(table_get_strs(addrTable, "address", "*", NULL));
            std::string mask(table_get_strs(addrTable, "mask", "255.255.255.255", NULL));
            allowedSubnets.push_back(Subnet(addr, mask));
        }
    }
    catch (MasException& e)
    {
        table_free_iter(iter);
        throw e;
    }
    catch (...)
    {
        table_free_iter(iter);
        throw MasException("Error creating sockets");
    }
    table_free_iter(iter);
}


void Httpd::createHandlers(HTable conf)
{
    if (! conf) return;
    HTableIterator iter = table_get_iter(conf);
    try
    {
        while (iter && table_iter_next(iter))
        {
            HTable handlerTable = table_iter_get_table(iter, NULL);
            std::string type(table_get_strs(handlerTable, "type", "", NULL));
            std::string location(table_get_strs(handlerTable, "location", "", NULL));
            addHandler(location, createHandler(type, handlerTable));
        }
    }
    catch (MasException& e)
    {
        table_free_iter(iter);
        throw e;
    }
    catch (...)
    {
        table_free_iter(iter);
        throw MasException("Error creating httpd handlers");
    }
    table_free_iter(iter);
}


void Httpd::closeSockets()
{
    for (std::vector<int>::iterator i = sockets.begin(); i != sockets.end(); i++)
    {
        int sock = *i;
        shutdown(sock, 2);
        close(sock);
    }
}


int Httpd::addSockets(fd_set *rdfs)
{
    int maxSock = 0, sock;
    
    for (std::vector<int>::iterator i = sockets.begin(); i != sockets.end(); i++)
    {
        sock = *i;
        FD_SET(sock, rdfs);
        if (sock > maxSock)
            maxSock = sock;
    }

    return maxSock;
}


bool Httpd::isValidIncomingAddress(unsigned long int addr)
{
    for (std::vector<Subnet>::iterator i = allowedSubnets.begin(); 
            i != allowedSubnets.end(); i++)
    {
        if ((*i).containsAddress(addr))
            return true;
    }
    return false;
}


void Httpd::acceptSocket(int sock)
{
    struct sockaddr_in clntAddr;
    memset(&clntAddr, 0, sizeof(clntAddr));
#ifdef WIN32
    int addrlen = sizeof(clntAddr);
#else
    socklen_t addrlen = sizeof(clntAddr);
#endif
    int connCnt;
    
    threadsLock.lock();
    connCnt = connections.size();
    threadsLock.unlock();
            
    int clntSock = accept(sock, (struct sockaddr*)&clntAddr, &addrlen);    
    if (clntSock < 0)
    {
        masLog.warn("Error accepting socket");
        return;
    }
    if (connCnt >= maxConnections)
    {
        shutdown(clntSock, 2);
        close(clntSock);
        masLog.warn("Connection from " + inet_atos(clntAddr.sin_addr.s_addr) + 
                " refused: too many connections");
        return;
    }
    if (! isValidIncomingAddress(clntAddr.sin_addr.s_addr))
    {
        shutdown(clntSock, 2);
        close(clntSock);
        masLog.warn("Refused connection from " + inet_atos(clntAddr.sin_addr.s_addr));
        return;
    }

#ifdef WIN32
    unsigned long int block = 1;
    ioctlsocket(clntSock, FIONBIO, &block);
#else
    fcntl(clntSock, F_SETFL, O_NONBLOCK);
#endif

    HttpdConnection *conn;
    try 
    {
        conn = new HttpdConnection(clntSock, this);
    }
    catch (MasException& e)
    {
        shutdown(clntSock, 2);
        close(clntSock);
        masLog.error(e.getMessage());
        return;
    }
    catch (...)
    {
        shutdown(clntSock, 2);
        close(clntSock);
        masLog.error("Error creating connection");
        return;
    }

    //Server::getServer()->getThreadPool()->addJob(conn, NULL);
    conn->start(false, true);
}


void Httpd::acceptSockets(fd_set *rdfs)
{
    int sock;
    
    for (std::vector<int>::iterator i = sockets.begin(); i != sockets.end(); i++)
    {
        sock = *i;
        if (FD_ISSET(sock, rdfs))
            acceptSocket(sock);
    }
}


void Httpd::run()
{
    fd_set rfds;
    int maxSocket, res;
    struct timeval tv;

    while (! isShuttingDown())
    {
        FD_ZERO(&rfds);
        maxSocket = addSockets(&rfds);
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        res = select(maxSocket + 1, &rfds, NULL, NULL, &tv);
        if (res > 0)
            acceptSockets(&rfds);
    }
    
    {
        Locker locker(threadsLock);
        for (std::list<HttpdConnection*>::iterator i = connections.begin(); 
                i != connections.end(); i++)
        {
            if (*i)
                (*i)->stop(false);
        }
        while (connections.size() > 0)
            stopCond.wait(threadsLock);
    }
}


void Httpd::addConnection(HttpdConnection *conn)
{
    Locker locker(threadsLock);
    connections.push_back(conn);
}


void Httpd::removeConnection(HttpdConnection *conn)
{
    Locker locker(threadsLock);
    connections.remove(conn);
    stopCond.signal();
}


void Httpd::processMimeLine(const std::string& line)
{
    int i = 0;    
    int len = line.length(), extStart;
    
    while ((i < len) && (line[i] != ' ') && (line[i] != '\t'))
    {
        if ((line[i] == '#') || (line[i] == '\n') || (line[i] == '\r'))
            return;
        i++;
    }
    
    std::string mimeType = line.substr(0, i);
    while (i < len)
    {
        while ((i < len) && ((line[i] == ' ') || (line[i] == '\t')))
            i++;
        if ((i >= len) || (line[i] == '#') || (line[i] == '\n') || (line[i] == '\r'))
            return;
        extStart = i;
        while ((line[i] != ' ') && (line[i] != '\t') && (line[i] != '#') 
                && (line[i] != '\n') && (line[i] != '\r') && (i < len))
            i++;
        if (i - extStart > 0)
        {
            std::string ext(toLowerCase(line.substr(extStart, i - extStart)));
            mimeExtensions[ext] = mimeType;
        }
        if ((i >= len) || (line[i] == '#') || (line[i] == '\n') || (line[i] == '\r'))
            return;
    }
}


void Httpd::loadMimeTypes(const std::string& fileName)
{
    std::ifstream ifs(fileName.c_str());
    if (! ifs.good())
        throw MasException("Error opening mime config '" + fileName + "'");
    mimeExtensions.clear();
    std::string s;
    while (std::getline(ifs, s))
    {
        if (s != "")
            processMimeLine(s);
    }
    ifs.close();
}


std::string& Httpd::getMimeType(const std::string& extension)
{
    Locker locker(mimeLock);
    std::string& type = mimeExtensions[toLowerCase(extension)];
    if (type == "")
        type = mimeExtensions[""];
    return type;
}


HttpdHandler* Httpd::createHandler(const std::string& type, HTable conf)
{
    if (type == "filetransfer")
        return new FileTransferHandler(conf);
    else
        throw MasException("Unknown handler type '" + type + "'");
}


HttpdHandler* Httpd::getHandler(const std::string& path)
{
    Locker locker(handlersLock);
    if (path == "")
    {
        for (std::set<HandlerRecord, HandlerComparator>::iterator i = 
                handlers.begin(); i != handlers.end(); i++)
        {
            const HandlerRecord& hrec = (*i);
            if (hrec.url == "")
                return hrec.handler;
        }
    }
    else
    {
        int len = path.length();
        for (std::set<HandlerRecord, HandlerComparator>::iterator i = 
                handlers.begin(); i != handlers.end(); i++)
        {
            const HandlerRecord& hrec = (*i);
            int ulen = hrec.url.length();
            if (! ulen)
                return hrec.handler;
            else
            {
                if (ulen <= len)
                {
                    if (! strncmp(hrec.url.c_str(), path.c_str(), hrec.url.length()))
                        return hrec.handler;
                }
            }
        }
    }
    return NULL;
}



extern "C" int EXPORT_REGFUNC register_httpd_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addRunnablePlugin("httpd", new Httpd());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

