#ifndef __HTTPD_H__
#define __HTTPD_H__


#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>
#include "conf.h"
#include "exceptions.h"
#include "threadpool.h"
#include "plugin.h"
#include "thread.h"


#ifdef WIN32
#ifdef BUILDING_HTTPDDLL
#define HTTPDIMPORT __declspec (dllexport)
#else
#define HTTPDIMPORT __declspec (dllimport)
#endif
#else
#define HTTPDIMPORT
#endif

class HTTPDIMPORT HttpException: public MasException
{
    private:
        int number;

    public:
        HttpException(int httpNumber, const std::string& msg);
    
    public:
        int getHttpCode() const;
};


class HTTPDIMPORT ServerURL
{
    private:
        std::string unparsed;
        std::string unparsedOptions;
        std::string path;
        std::map<std::string, std::string> options;
    
    public:
        ServerURL(const std::string& url);
        ~ServerURL();

    public:
        const std::string& getOptions() const;
        const std::string& getOption(const std::string& option, const std::string& dflt);
        const int getOption(const std::string& option, int dflt);
        const double getOption(const std::string& option, double dflt);
        static std::string unescapeStr(const std::string& str);
        const std::string& getPath() const;

    private:
        void parseOptions();
};


class Httpd;


class HTTPDIMPORT HttpdConnection: public Thread
{
    private:
        std::list<std::string> sendHeaders;
        std::map<std::string, std::string> receivedHeaders;
        int sock;
        Httpd *httpd;
        unsigned char *readBuf;
        int bufFilled, bufPosition;
        std::string method;
        std::string protocol;
        std::string document;
        std::string status;
        bool sending;
        int contentLength;
    
    public:
        HttpdConnection(int s, Httpd *parent);
        virtual ~HttpdConnection();

    public:
        virtual void run();
        void sendData(const unsigned char *data, int length, int timeout=60);
        void sendData(const std::string& str, int timeout=60);
        void transferHeaders();
        void setContentType(const std::string& ctype);
        void setHttpHeader(const std::string& header, const std::string& value);
        Httpd* getHttpd();
        void setStatus(int code);
        const std::string& getHeader(const std::string& name);
        void setContentLength(int length);

    private:
        std::string readLine();
        int readChar();
        void unreadChar(int ch);
        void parseHeader();
        void parseMethod(const std::string& methodStr);
        void parseHeaderLine(const std::string& line);
        void processRequest();
        void showErrorMessage(int code, const std::string& msg);
};



enum RequestType
{
    REQ_GET,
    REQ_POST,
    REQ_HEAD
};
        


class HTTPDIMPORT HttpdHandler
{
    public:
        HttpdHandler() { };
        virtual ~HttpdHandler() { };
        virtual int handleRequest(HttpdConnection& connection, const ServerURL& url, 
                enum RequestType requestType) = 0;
};



class HTTPDIMPORT Subnet
{
    private:
        unsigned long int addr;
        unsigned long int mask;

    public:
        Subnet(const std::string& a, const std::string& m);
        Subnet(const std::string& a);
        Subnet(const Subnet& subnet);
        ~Subnet();

    public:
        unsigned long int getAddress() const;
        unsigned long int getMask() const;
        bool containsAddress(unsigned long int a);
        bool containsAddress(const std::string& a);
        std::string toString();

    public:
        static unsigned long int atoip(const std::string& str);
        static unsigned long int atoipmask(const std::string& str);
};



class HTTPDIMPORT Httpd: public RunnablePlugin
{
    private:
        class HandlerRecord
        {
            public:
            HttpdHandler* handler;
            std::string url;
            HandlerRecord(const std::string& url, HttpdHandler *handler) {
                this->url = url;
                this->handler = handler;
            };
            HandlerRecord(const HandlerRecord& rec) {
                url = rec.url;
                handler = rec.handler;
            };
            bool operator < (const HandlerRecord& rec) const {
                return url.length() < rec.url.length();
            };
            bool operator > (const HandlerRecord& rec) const {
                return url.length() > rec.url.length();
            };
            bool operator == (const HandlerRecord& rec) const {
                return url == rec.url;
            };
            void operator = (const HandlerRecord& rec) {
                if (this != &rec)
                {
                    url = rec.url;
                    handler = rec.handler;
                }
            };
        };
        
        struct HandlerComparator /*: public std::binary_function<const HandlerRecord& h1, const HandlerRecord& h2, bool> */
        {
            bool operator()(const HandlerRecord& h1, const HandlerRecord& h2)
            {
                return h1 > h2;
            };
        };
        
        //std::set<HandlerRecord, std::greater<HandlerRecord> > handlers;
        std::set<HandlerRecord, HandlerComparator> handlers;
        std::vector<int> sockets;
        std::vector<Subnet> allowedSubnets;
        std::list<HttpdConnection*> connections;
        int maxConnections;
        Lock threadsLock;
        Lock stopLock;
        Cond stopCond;
        std::map<std::string, std::string> mimeExtensions;
        Lock mimeLock;
        Lock handlersLock;
    
    public:
        Httpd();
        virtual ~Httpd();

    public:
        void addHandler(const std::string& url, HttpdHandler *handler); 
        void removeHandler(const std::string& url); 
        virtual void run();
        void addConnection(HttpdConnection *conn);
        void removeConnection(HttpdConnection *conn);
        std::string& getMimeType(const std::string& extension);
        HttpdHandler* getHandler(const std::string& path);

    protected:
        virtual HttpdHandler* createHandler(const std::string& type, HTable conf);
        
    private:
        void createSockets(HTable listenConf);
        void createAllowed(HTable allowedConf);
        void createHandlers(HTable handlersConf);
        void closeSockets();
        int addSockets(fd_set *rdfs);
        void acceptSockets(fd_set *rdfs);
        void acceptSocket(int sock);
        bool isValidIncomingAddress(unsigned long int addr);
        //bool isServerStopping();
        void processMimeLine(const std::string& line);
        void loadMimeTypes(const std::string& fileName);
};


class HTTPDIMPORT FileTransferHandler: public HttpdHandler
{
    private:
        std::string location;
        std::string directory;
        std::vector<std::string> indexes;
    
    public:
        FileTransferHandler(HTable conf);
        virtual ~FileTransferHandler();
 
    public:
        virtual int handleRequest(HttpdConnection& connection, const ServerURL& url, 
                enum RequestType requestType);

    private:
        void buildIndexes(HTable table);
};


#endif

