#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#ifdef WIN32
#include <windows.h>
#else
#include <sys/socket.h>
#endif
#include "httpd.h"
#include "exceptions.h"
#include "utils.h"
#include "log.h"


#define READ_BUF_SIZE 2048
#define MAX_LINE_SIZE 20480


HttpdConnection::HttpdConnection(int s, Httpd* parent)
{
    sock = s;
    httpd = parent;
    readBuf = new unsigned char[READ_BUF_SIZE];
    bufFilled = 0;
    bufPosition = 0;
}


HttpdConnection::~HttpdConnection()
{
    getHttpd()->removeConnection(this);
    delete[] readBuf;
    shutdown(sock, 2);
    close(sock);
}


void HttpdConnection::run()
{
    while (! isShuttingDown())
    {
        sendHeaders.clear();
        sending = false;
        contentLength = -1;
        try
        {
            parseHeader();
            processRequest();
            if (contentLength < 0)
                break;
        }
        catch (HttpException& he)
        {
            if (! sending)
                showErrorMessage(he.getHttpCode(), he.getMessage());
            masLog.warn(he.getMessage());
            return;
        }
        catch (MasException& me)
        {
            if (! sending)
                showErrorMessage(500, me.getMessage());
            masLog.warn(me.getMessage());
            return;
        }
        catch (...)
        {
            if (! sending)
                showErrorMessage(500, "Unknown exception");
            masLog.warn("Unknown exception during http connection");
            return;
        }
    }
}


static RequestType getRequestMethod(const std::string& s)
{
    if (s == "get")
        return REQ_GET;
    else if (s == "post")
        return REQ_POST;
    else if (s == "head")
        return REQ_POST;
    else
        throw HttpException(501, "Method not implemented");
}


void HttpdConnection::processRequest()
{
    HttpdHandler *handler = httpd->getHandler(document);
    if (! handler)
        throw HttpException(404, "Handler not found");
    else
    {
        RequestType reqType = getRequestMethod(toLowerCase(method));
        handler->handleRequest(*this, ServerURL(document), reqType);
    }
}


static struct
{
    int code;
    std::string msg;
} httpMessages[] = { 
      { 100, "Continue" },
      { 101, "Switching protocols" },
      { 200, "OK" },
      { 201, "Created" },
      { 202, "Accepted" },
      { 203, "Non-Authoritative Information" },
      { 204, "No Content" },
      { 205, "Reset Content" },
      { 206, "Partial Content" },
      { 300, "Multiple Choices" },
      { 301, "Moved Permanently" },
      { 302, "Moved Temporariry" },
      { 303, "See Other" },
      { 304, "Not Modified" },
      { 305, "Use Proxy" },
      { 400, "Bad Request" },
      { 401, "Unauthorized" },
      { 402, "Payment required" },
      { 403, "Forbidden" },
      { 404, "Not Found" },
      { 405, "Method Not Allowed" },
      { 406, "Not Acceptable" },
      { 407, "Proxy Authorization Required" },
      { 408, "Request Time-out" },
      { 409, "Conflict" },
      { 410, "Gone" },
      { 411, "Length Required" },
      { 412, "Precondition Failed" },
      { 413, "Request Entity Too Large" },
      { 414, "Request-URI Too Long" },
      { 415, "Unsupported Media Type" },
      { 500, "Internal Server Error" },
      { 501, "Not Implemented" },
      { 502, "Bad Gateway" },
      { 503, "Service Unavailable" },
      { 504, "Gateway Time-out" },
      { 505, "HTTP Version not supported" },
      { 0, "" }
    };


static const std::string& getMessageDescr(int number)
{    
    int i = 0;
    while (httpMessages[i].code != 0)
    {
        if (httpMessages[i].code == number)
            return httpMessages[i].msg;
        i++;
    }
    static std::string unknown("Unknown error code");
    return unknown;
}


void HttpdConnection::showErrorMessage(int number, const std::string& msg)
{
    sendHeaders.erase(sendHeaders.begin(), sendHeaders.end());
    setStatus(number);
    setContentType("text/html");
    transferHeaders();
    sendData("<html><header><title>Error</title></header><body><h1>" + 
            numToStr(number) + "</h1><p>" + msg + "</p></body></html>");
}


void HttpdConnection::setStatus(int code)
{
    status = "HTTP/1.0 " + numToStr(code) + " " + getMessageDescr(code);
}


void HttpdConnection::transferHeaders()
{
    sending = true;
    sendData(status + "\r\n");
    
    for (std::list<std::string>::iterator i = sendHeaders.begin(); 
            i != sendHeaders.end(); i++)
    {
        sendData(*i + "\r\n");
    }
    
    sendData("\n");
}


static void waitForWriteData(int sock, int timeout)
{
    fd_set wfds;
    int retval, attempt = 0;
    struct timeval tv;
    
    time_t start = time(NULL);
    while (1)
    {
        FD_ZERO(&wfds);
        FD_SET(sock, &wfds);
        memset(&tv, 0, sizeof(tv));
        tv.tv_sec = timeout;
        retval = select(sock + 1, NULL, &wfds, NULL, NULL);
        if (retval > 0)
            return;
        attempt++;
        if (time(NULL) - start > timeout)
            return;
        if (attempt > 5)
#ifdef WIN32
            Sleep(1000);
#else
            usleep(1000000);
#endif
    }
}


void HttpdConnection::sendData(const unsigned char *data, int length, int timeout)
{
    int sent, pos = 0;
    
    if (length < 1)
        return;
    do
    {
        waitForWriteData(sock, timeout);
#ifndef WIN32
        sent = send(sock, data + pos, length, 0);
#else
        sent = send(sock, (char*)data + pos, length, 0);
#endif
        if (sent < 0)
            throw MasException("Error sending http data");
        pos += sent;
        length -= sent;
    } while (length > 0);
}


void HttpdConnection::sendData(const std::string& str, int timeout)
{
    sendData((const unsigned char*)str.c_str(), str.length(), timeout);
}


void HttpdConnection::setContentType(const std::string& ctype)
{
    setHttpHeader("Content-type", ctype);
}


void HttpdConnection::setHttpHeader(const std::string& header, const std::string& value)
{
    sendHeaders.push_back(header + ": " + value);
}


Httpd* HttpdConnection::getHttpd()
{
    return httpd;
}


/*static bool hasData(int sock)
{
    fd_set rfds;
    FD_ZERO(&rfds);
    FD_SET(sock, &rfds);
    struct timeval tv;
    memset(&tv, 0, sizeof(tv));
    int retval = select(sock + 1, &rfds, NULL, NULL, &tv);
    if (retval > 0)
        return true;
    else
        return false;
}*/


static void waitForData(int sock, int timeout=60)
{
    fd_set rfds;
    int retval, attempt = 0;
    struct timeval tv;
    
    time_t start = time(NULL);
    while (1)
    {
        FD_ZERO(&rfds);
        FD_SET(sock, &rfds);
        memset(&tv, 0, sizeof(tv));
        tv.tv_sec = timeout;
        retval = select(sock + 1, &rfds, NULL, NULL, &tv);
        if (retval > 0)
            return;
        if (time(NULL) - start > timeout)
            return;
        attempt++;
        if (attempt > 5)
#ifndef WIN32
            usleep(1000000);
#else
            Sleep(1000);
#endif
    }
}


static int recvToBuf(int sock, unsigned char *buf, int maxLength)
{
#ifndef WIN32
    return recv(sock, buf, maxLength, 0);
#else
    return recv(sock, (char*)buf, maxLength, 0);
#endif
}


int HttpdConnection::readChar()
{
    if (bufPosition >= bufFilled)
    {
        waitForData(sock);
        bufPosition = 0;
        bufFilled = recvToBuf(sock, readBuf, READ_BUF_SIZE);
    }

    if (bufPosition >= bufFilled)
        return -1;
    else
        return readBuf[bufPosition++];
}


void HttpdConnection::unreadChar(int ch)
{
    if (ch == -1)
        return;
    if (bufPosition > 0)
        bufPosition--;
}


std::string HttpdConnection::readLine()
{
    std::string str;
    int ch;
    int size = 0;
    char sb[2];
    sb[1] = 0;

    ch = readChar();
    while ((ch != -1) && (ch != '\r'))
    {
        sb[0] = ch;
        str.append(sb);
        size++;
        if (size > MAX_LINE_SIZE)
            throw HttpException(413, "Header line too long");
        ch = readChar();
    }
    if (ch == '\r')
    {
        ch = readChar();
        if (ch != '\n')
            unreadChar(ch);
    }
    return str;
}


void HttpdConnection::parseMethod(const std::string& methodStr)
{
    int i = 0;
    int len = methodStr.length();
    int docStart = -1, verStart = -1;
    
    while (i < len)
    {
        if (methodStr[i] == ' ')
        {
            if (docStart < 0)
            {
                method = methodStr.substr(0, i);
                docStart = i + 1;
            }
            else
                if (verStart < 0)
                {
                    document = methodStr.substr(docStart, i - docStart);
                    verStart = i + 1;
                }
        }
        i++;
    }
    if (verStart > 0)
        protocol = methodStr.substr(verStart);

    if ((method == "") || (protocol == "") || (document == ""))
        throw HttpException(400, "Invalid method line '" + methodStr + "'");
}


void HttpdConnection::parseHeaderLine(const std::string& line)
{
    int i = 0;
    int len = line.length();
    
    while (i < len)
    {
        if (line[i] == ':')
        {
            int nameEnd = i;
            i++;
            while ((i < len) && ((line[i] == ' ') || (line[i] == '\t')))
                i++;
            std::string name = toLowerCase(line.substr(0, nameEnd));
            std::string value = line.substr(i);
            receivedHeaders[name] = value;
            break;
        }
        i++;
    }
}


const std::string& HttpdConnection::getHeader(const std::string& line)
{
    std::string name = toLowerCase(line);
    return receivedHeaders[name];
}


void HttpdConnection::parseHeader()
{
    std::string s;
    
    s = readLine();
    if (s == "")
        throw HttpException(400, "Error reading method line");
    parseMethod(s);

    s = readLine();
    while (s != "")
    {
        parseHeaderLine(s);
        s = readLine();
    }
}

        
void HttpdConnection::setContentLength(int length)
{
    contentLength = length;
    setHttpHeader("Content-Length", numToStr(length));
}

