#include <iostream>
#include <fstream>
#include "arena.h"
#include "exceptions.h"
#include "server.h"
#include "log.h"
#include "utils.h"


Arena::Arena(const std::string &fname)
{
    filename = fname;
    load();
}


Arena::~Arena()
{
    Unit *u;
    
    for (std::map<std::string, Unit*>::iterator i = units.begin(); 
            i != units.end(); i++)
    {
        u = (*i).second;
        if (u)
        {
            if (u->isRunning())
                u->stop();
        }
    }

    for (std::map<std::string, Unit*>::iterator i = units.begin(); 
            i != units.end(); i++)
    {
        u = (*i).second;
        if (u)
        {
            if (u->hasOutputs())
            {
                Link *l = u->getOutputLink();
                if (l) delete l;
            }
            delete u;
        }
    }
}


static Unit* createUnit(const std::string &type, HTable conf)
{
    Unit *unit = NULL;
    try
    {
        unit = Server::getServer()->createUnit(type, conf);
    }
    catch (MasException &e)
    {
        masLog.error("Error creating unit '" + type + "': " + e.getMessage());
    }
    catch (...)
    {
        masLog.error("Error creating unit '" + type + "'");
    }
    return unit;
}


void Arena::loadUnits(HTable conf)
{
    HTableIterator iter = table_get_iter(table_get_table(conf, "units", 
                NULL, NULL));
    if (iter)
    {
        HTable unitTable;
        while (table_iter_next(iter))
        {
            unitTable = table_iter_get_table(iter, NULL);
            if (unitTable)
            {
                std::string type(table_get_strs(unitTable, "type", "", NULL));
                if (type != "")
                {
                    Unit *unit = createUnit(type, unitTable);
                    if (unit)
                        addUnit(unit);
                }
                else
                    masLog.error("Missing unit type");
            }
            else
                masLog.error("Invalid unit entry in streams");
        }
        table_free_iter(iter);
    }
}


void Arena::loadLinks(HTable conf)
{
    HTableIterator iter = table_get_iter(table_get_table(conf, "links", 
                NULL, NULL));
    if (iter)
    {
        HTable linkTable;
        while (table_iter_next(iter))
        {
            linkTable = table_iter_get_table(iter, NULL);
            if (linkTable)
            {
                std::string fromName(table_get_strs(linkTable, "from", "", NULL));
                std::string toName(table_get_strs(linkTable, "to", "", NULL));
                if ((fromName != "") && (toName != ""))
                {
                    try
                    {
                        addLink(fromName, toName);
                    }
                    catch (MasException &e)
                    {
                        masLog.error("Error creating link '" + fromName + "' -> '" +
                                toName + "': " + e.getMessage());
                    }
                    catch (...)
                    {
                        masLog.error("Error creating link '" + fromName + "' -> '" +
                                toName + "'");
                    }
                }
                else
                    masLog.error("Invalid link description");
            }
            else
                masLog.warn("Invalid link entry in streams");
        }
        table_free_iter(iter);
    }
}


void Arena::load()
{
    HTable streams = table_read(filename.c_str());
    if (! streams)
        throw MasException("Error reading streams file '" + filename + "'");

    loadUnits(streams);
    loadLinks(streams);

    table_free(streams);


    for (std::map<std::string, Unit*>::iterator i = units.begin();
            i != units.end(); i++)
    {
        try
        {
            if ((*i).second)
                (*i).second->start();
            else
                masLog.error("Invalid unit in arena");
        }
        catch (MasException &e)
        {
            masLog.error("Error starting unit '" + (*i).first + "': " +
                    e.getMessage());
        }
        catch (...)
        {
            masLog.error("Error starting unit '" + (*i).first + "'");
        }
    }
}


void Arena::save()
{
    HTable conf = table_create();
    if (! conf)
        throw MasException("Error creating streams table");
    
    HTable unitsTable = table_create();
    if (! unitsTable)
    {
        table_free(conf);
        throw MasException("Error creating units table");
    }
    table_set_table(conf, "units", unitsTable);

    for (std::map<std::string, Unit*>::iterator i = units.begin(); 
            i != units.end(); i++)
    {
        Unit *u = (*i).second;
        HTable unitTable = table_create();
        if (! unitTable)
        {
            table_free(conf);
            throw MasException("Error creating unit table");
        }
        table_append_table(unitsTable, unitTable);
        u->save(unitTable);
    }

    HTable linksTable = table_create();
    if (! linksTable)
    {
        table_free(conf);
        throw MasException("Error creating links table");
    }
    table_set_table(conf, "links", linksTable);
    
    for (std::set<Chain>::iterator i = chains.begin(); i != chains.end(); i++)
    {
        HTable linkTable = table_create();
        if (! linkTable)
        {
            table_free(conf);
            throw MasException("Error creating link table");
        }
        table_append_table(linksTable, linkTable);
        table_set_str(linkTable, "from", (*i).from->getName().c_str());
        table_set_str(linkTable, "to", (*i).to->getName().c_str());
    }

    char *s = table_to_str(conf, 1, 1, 1);
    table_free(conf);
    if (! s)
        throw MasException("Error converting streams table to text");
    
    std::ofstream of(filename.c_str(), std::ios::trunc);
    if (! of.good())
    {
        free(s);
        throw MasException("Error opening streams file for writing");
    }
    of << s << std::endl;
    free(s);
    of.close();
}


void Arena::addUnit(Unit *unit)
{
    if (! unit)
        throw MasException("Attept of adding invalid unit");
    if (units[unit->getName()])
        throw MasException("Unit already exists");
    units[unit->getName()] = unit;
}


void Arena::delUnit(Unit *unit)
{
    if (! unit)
        throw MasException("Attept of deleting invalid unit");
    
    std::list<Chain> toRemove;
    for (std::set<Chain>::iterator i = chains.begin(); i != chains.end(); i++)
        if (((*i).from == unit) || ((*i).to == unit))
            toRemove.push_back(*i);
    for (std::list<Chain>::iterator i = toRemove.begin(); i != toRemove.end(); i++)
        delLink((*i).from, (*i).to);
    
    units.erase(unit->getName());
}


void Arena::addLink(const std::string &fromName, const std::string &toName)
{
    if (units.count(fromName) < 1)
        throw MasException("Invalid input unit '" + fromName + "'");
    if (units.count(toName) < 1)
        throw MasException("Invalid input unit '" + toName + "'");
    Unit *from = units[fromName];
    Unit *to = units[toName];
    addLink(from, to);
}


void Arena::addLink(Unit *from, Unit *to)
{
    if ((! from) || (! to))
        throw MasException("Invalid units");

    Chain c(from, to);
    if (chains.count(c) > 0)
        throw MasException("Link already exists");
    
    if (! from->hasOutputs())
        throw MasException("Source unit hasn't outputs");
    if (! to->hasInputs())
        throw MasException("Destination unit hasn't inputs");
    if (to->getMaxInputs() != -1)
        if (to->getInputsCount() >= to->getMaxInputs())
            throw MasException("Destination unit hasn't free input slots");
    
    Link *link = from->getOutputLink();
    if (! link)
    {
        link = new Link(from->getOutputBufferSize());
        from->setOutputLink(link);
    }
    to->addInputLink(link);
    chains.insert(c);
}


void Arena::delLink(Unit *from, Unit *to)
{
    if ((! from) || (! to))
        throw MasException("Invalid units");

    Chain c(from, to);
    if (chains.count(c) < 1)
        throw MasException("Link doesn't exists");

    Link *link = from->getOutputLink();
    if (! link)
        throw MasException("Source unit hasn't output link");
    to->delInputLink(link);

    chains.erase(c);

    int refCnt = 0;
    for (std::set<Chain>::iterator i = chains.begin(); i != chains.end(); i++)
        if ((*i).from == from)
        {
            refCnt++;
            break;
        }

    if (! refCnt)
    {
        from->removeOutputLink();
        delete link;
    }
}


std::string Arena::genUniqueName(const std::string &base)
{
    std::string s;

    for (int i = 1; i < 100000; i++)
    {
        s = base + numToStr(i);
        if (units.count(s) < 1)
            return s;
    }

    throw MasException("Error creating unique unit name");
}

