#include "link.h"
#include "exceptions.h"
#include "log.h"


Link::Link(int maxBufSize, int maxPacketsCount)
{
    packetsDataBufferSize = maxBufSize;
    packetsInfoBufferSize = maxPacketsCount;
    packets = new Packet[maxPacketsCount];
    buffer = new unsigned char[maxBufSize];
    firstPacketIndex = nextPacketIndex = 0;
    firstPacketNo = 0;
    packetsCount = 0;
    lastPacketMediaType = NULL;
    lastPacketMediaInfo = NULL;
}


Link::~Link()
{
    for (std::list<PacketMediaType*>::iterator i = packetsMediaTypes.begin();
            i != packetsMediaTypes.end(); i++)
    {
        if (*i)
        {
            if ((*i)->mtype)
                delete (*i)->mtype;
            delete *i;
        }
    }
    
    for (std::list<PacketMediaInfo*>::iterator i = packetsMediaInfos.begin();
            i != packetsMediaInfos.end(); i++)
    {
        if (*i)
        {
            if ((*i)->minfo)
                delete (*i)->minfo;
            delete *i;
        }
    }
    
    delete[] packets;
    delete[] buffer;
}


void Link::removeFirstPacket()
{
    if (packetsCount > 0)
    {
        PacketMediaType *pmt;
        pmt = packets[firstPacketIndex].packetMediaType;
        if (pmt)
        {
            pmt->refCount--;
            if ((pmt->refCount < 1) && (pmt != lastPacketMediaType))
            {
                packetsMediaTypes.remove(pmt);
                delete pmt->mtype;
                delete pmt;
            }
            //packets[firstPacketIndex].packetMediaType = NULL;
        }
        
        PacketMediaInfo *pmi;
        pmi = packets[firstPacketIndex].packetMediaInfo;
        if (pmi)
        {
            pmi->refCount--;
            if ((pmi->refCount < 1) && (pmi != lastPacketMediaInfo))
            {
                packetsMediaInfos.remove(pmi);
                delete pmi->minfo;
                delete pmi;
            }
            //packets[firstPacketIndex].packetMediaInfo = NULL;
        }
        
        firstPacketIndex++;
        firstPacketNo++;
        packetsCount--;
        if (firstPacketIndex >= packetsInfoBufferSize)
            firstPacketIndex -= packetsInfoBufferSize;        
    }
}

void Link::setEmptyMediaInfo()
{
    PacketMediaInfo *pmi = new PacketMediaInfo;
    pmi->refCount = 0;
    pmi->minfo = new MediaInfo();
    packetsMediaInfos.push_back(pmi);
    lastPacketMediaInfo = pmi;
}


void Link::setMediaInfo(const MediaInfo &mediaInfo)
{
    Locker locker(lockBuffer);
    PacketMediaInfo *pmi = new PacketMediaInfo;
    pmi->refCount = 0;
    pmi->minfo = new MediaInfo(mediaInfo);
    packetsMediaInfos.push_back(pmi);
    lastPacketMediaInfo = pmi;
}


void Link::addData(unsigned char *data, int length, double playTime, const MediaType &mtype)
{
    Locker locker(lockBuffer);

    if ((! data) || (length < 1) || (length > packetsDataBufferSize))
        // throw exception?
        return;

    int lastPacketIndex = nextPacketIndex - 1;
    if (lastPacketIndex < 0)
        lastPacketIndex += packetsInfoBufferSize;
    
    // check if we have enough room in data buffer
    while (packetsCount > 0)
    {
        int firstPacketStart = packets[firstPacketIndex].start;
        int lastPacketStart = packets[lastPacketIndex].start;
        int filled;
        if (lastPacketStart < firstPacketStart)
            filled = packetsDataBufferSize - firstPacketStart + lastPacketStart;
        else
            filled = lastPacketStart - firstPacketStart;
        filled += packets[lastPacketIndex].length;
        if (packetsDataBufferSize - filled >= length)
            break;
        else
            removeFirstPacket();
    }
    
    // check if we have room in packets buffer
    while ((packetsCount > 0) && (packetsCount >= packetsInfoBufferSize))
        removeFirstPacket();
   
    int lastPacketEnd;
    // always start at begin of the buffer.  just in case.
    if (packetsCount < 1)
        firstPacketIndex = nextPacketIndex = lastPacketEnd = 0;
    else
    {
        lastPacketEnd = packets[lastPacketIndex].start + packets[lastPacketIndex].length;
        if (lastPacketEnd > packetsDataBufferSize)
            lastPacketEnd -= packetsDataBufferSize;
    }
 
    // fill packet info
    packets[nextPacketIndex].start = lastPacketEnd;
    packets[nextPacketIndex].length = length;
    if (playTime < 0)
        playTime = 0;
    packets[nextPacketIndex].playTime = playTime;
    
    // set packet media type
    if (lastPacketMediaType)
    {
        if (! lastPacketMediaType->mtype->isEqual(mtype))
        {
            if (lastPacketMediaType->refCount < 1)
            {
                packetsMediaTypes.remove(lastPacketMediaType);
                delete lastPacketMediaType->mtype;
                delete lastPacketMediaType;
                lastPacketMediaType = NULL;
            }
        }
    }
    if (! lastPacketMediaType)
    {
        lastPacketMediaType = new PacketMediaType;
        lastPacketMediaType->mtype = mtype.clone();
        lastPacketMediaType->refCount = 0;
        packetsMediaTypes.push_back(lastPacketMediaType);
    }
    packets[nextPacketIndex].packetMediaType = lastPacketMediaType;
    lastPacketMediaType->refCount++;
    
    // set packet media info
    if (! lastPacketMediaInfo)
        setEmptyMediaInfo();
    packets[nextPacketIndex].packetMediaInfo = lastPacketMediaInfo;
    lastPacketMediaInfo->refCount++;
    
    // copy data to buffer
    if (packetsDataBufferSize - lastPacketEnd >= length)
        memcpy(buffer + lastPacketEnd, data, length);
    else
    {
        int split = packetsDataBufferSize - lastPacketEnd;
        memcpy(buffer + lastPacketEnd, data, split);
        memcpy(buffer, data + split, length - split);
    }
    
    nextPacketIndex++;
    if (nextPacketIndex >= packetsInfoBufferSize)
        nextPacketIndex -= packetsInfoBufferSize;
    packetsCount++;

//std::cout << "addData, packets " << packetsCount << " " << firstPacketIndex << " " << nextPacketIndex << std::endl;
/*if (firstPacketIndex <= nextPacketIndex)
    std::cout << packetsInfoBufferSize - firstPacketIndex + nextPacketIndex;
else
    std::cout << nextPacketIndex - firstPacketIndex;*/
//std::cout << std::endl;
/*{
    FILE *f;
    char buf[100];
    sprintf(buf, "/tmp/mpnew/%i.frame.mp3", firstPacketNo + packetsCount);
    f = fopen(buf, "wb");
    fwrite(data, 1, length, f);
    fclose(f);
}*/
    
    notEmpty.signal();
}


void Link::beginGet()
{
    lockBuffer.lock();
}


int Link::waitForPacket(unsigned long long int &packetNumber)
{
    packetNumber = correctPacketNo(packetNumber);
    while ((packetsCount == 0) || (firstPacketNo + packetsCount <= packetNumber))
    {
        notEmpty.wait(lockBuffer);
        packetNumber = correctPacketNo(packetNumber);
    }
  
    // find requested packet
    int ourPacketIndex = packetNumber - firstPacketNo + firstPacketIndex;
    if (ourPacketIndex >= packetsInfoBufferSize)
        ourPacketIndex -= packetsInfoBufferSize;

    return ourPacketIndex;
}


void Link::getPacketInfo(MediaType* &mediaType, MediaInfo* &mediaInfo, 
        int &length, double &playTime, unsigned long long int &packetNo)
{
    int ourPacketIndex = waitForPacket(packetNo);
   
    playTime = packets[ourPacketIndex].playTime;
    length = packets[ourPacketIndex].length;
    mediaType = packets[ourPacketIndex].packetMediaType->mtype;
    mediaInfo = packets[ourPacketIndex].packetMediaInfo->minfo;
}


void Link::getData(unsigned long long int &packetNumber, unsigned char *data)
{
    if (! data)
        throw MasException("Invalid link buffer");

    int ourPacketIndex = waitForPacket(packetNumber);
   
    // copy data to buffer
    int dataStart = packets[ourPacketIndex].start;
    int length = packets[ourPacketIndex].length;
    if (dataStart + length <= packetsDataBufferSize)
        memcpy(data, buffer + dataStart, length);
    else
    {
        int split = packetsDataBufferSize - dataStart;
        memcpy(data, buffer + dataStart, split);
        memcpy(data + split, buffer, length - split);
    }

//std::cout << "getData, packet " << ourPacketIndex << std::endl;
    packetNumber++;
}

void Link::endGet()
{
    lockBuffer.unlock();
}


int Link::correctPacketNo(unsigned long long int packetNumber)
{
    // check if consumer packet number is ancient
    if (packetNumber < firstPacketNo)
    {
//if (packetNumber) std::cout << "skipping packet" << std::endl;
        packetNumber = firstPacketNo;
    }
    
    // check if consumer packet number is too big (overflow)
    if (packetNumber > firstPacketNo + packetsInfoBufferSize * 8)
        packetNumber = firstPacketNo;
        
    return packetNumber;
}
