#include <dirent.h>
#ifndef WIN32
#include <dlfcn.h>
#endif
#include <set>
#include <map>
#include <algorithm>
#include "plugin.h"
#include "exceptions.h"
#include "utils.h"
#include "log.h"
#include "server.h"



GenericLoader::~GenericLoader()
{
    ModuleHandler handler;
    for (std::vector<ModuleHandler>::iterator i = handlers.begin();
            i != handlers.end(); i++)
    {
        handler = *i;
#ifdef WIN32
        FreeLibrary(handler);
#else
        dlclose(handler);
#endif
    }
}


void GenericLoader::load(const std::string& path, const std::string& name, HTable descr)
{
#ifdef WIN32
    std::string filename(path + "\\" + name + ".dll");
#else
    std::string filename(path + "/" + name + ".so");
#endif
    ModuleHandler instance;
#ifdef WIN32
    instance = LoadLibrary(filename.c_str());
#else
    instance = dlopen(filename.c_str(), RTLD_NOW | RTLD_GLOBAL);
#endif
    if (instance)
    {
        std::string funcName = "register_" + name + "_plugins";
#ifdef WIN32
        int (__stdcall *reg_func)();
        reg_func = (int(__stdcall *)())GetProcAddress(instance, funcName.c_str());
#else
        int (*reg_func)();
        reg_func = (int(*)())dlsym(instance, funcName.c_str());
#endif
        if (! reg_func)
        {
#ifdef WIN32
            FreeLibrary(instance);
#else
            dlclose(instance);
#endif
            throw MasException("Module '" + name + "' has not register function");
        }
        if (reg_func())
        {
#ifdef WIN32
            FreeLibrary(instance);
#else
            dlclose(instance);
#endif
            throw MasException("Error registering module '" + name + "'");
        }
        handlers.push_back(instance);
    }
    else
    {
        char *msg = NULL;
#ifndef WIN32
        msg = (char*)dlerror();
#endif
        if (msg)
            throw MasException("Error loading module '" + filename + "': " + msg);
        else
            throw MasException("Error loading module '" + filename + "'");
    }
}



class Module
{
    private:
        std::set<std::string> depends;
        std::string loader;
        std::string path;
        std::string name;
        HTable conf;
    
    public:
        Module(const std::string& path, const std::string &filename);
        ~Module();

    public:
        std::string getLoader();
        std::string getPath();
        std::string getName();
        int depsCount();
        void removeDep(const std::string& name);
        HTable getConf();
};


Module::Module(const std::string& path, const std::string &filename)
{
    this->path = path;
#ifdef WIN32
    std::string fullname(path + "\\" + filename);
#else
    std::string fullname(path + "/" + filename);
#endif
    HTable table = table_read(fullname.c_str());
    if (! table)
        throw MasException("Error reading module '" + fullname + "'");
    loader = table_get_strs(table, "loader", "generic", NULL);
    name = table_get_strs(table, "name", "", NULL);
    if (name == "")
    {
        table_free(table);
        throw MasException("Invalid name in module '" + fullname + "'");
    }
    HTable depsTable = table_get_table(table, "depends", NULL, NULL);
    if (depsTable)
    {
        HTableIterator iter = table_get_iter(depsTable);
        if (iter)
        {
            std::string s;
            while (table_iter_next(iter))
            {
                s = std::string(table_iter_get_strs(iter, NULL));
                depends.insert(s);
            }
            table_free_iter(iter);
        }
    }
    conf = table;
}

Module::~Module()
{
    table_free(conf);
}

std::string Module::getLoader()
{
    return loader;
}

std::string Module::getName()
{
    return name;
}

std::string Module::getPath()
{
    return path;
}

int Module::depsCount()
{
    return depends.size();
}

void Module::removeDep(const std::string& name)
{
    depends.erase(name);
}

HTable Module::getConf()
{
    return conf;
}


static void addModulesFromDir(std::map<std::string, Module*> &modules, 
        const std::string &path)
{
    DIR *dir;
    std::string name, ext;
#ifndef WIN32
    struct dirent de, *nde;
#else
    struct dirent *nde;
#define de (*nde)
#endif

    dir = opendir(path.c_str());
    if (! dir)
        throw MasException("Error opening directory '" + path + "'");
    
#ifdef WIN32
    while ((nde = readdir(dir)))
#else
    while (! readdir_r(dir, &de, &nde))
#endif
    {
        if (! nde)
            break;
        try
        {
            if (de.d_name[0] != '.')
            {
                ext = getFileExtLower(de.d_name);
                if (ext == "plg")
                {
                    Module *module = new Module(path, de.d_name);
                    modules[module->getName()] = module;
                }
            }
        }
        catch (MasException &e)
        {
            masLog.warn(e.getMessage());
        }
        catch (...)
        {
            masLog.warn("Error processing module '" + name + "' from '" + path + "'");
        }
    }
    
    closedir(dir);
}


static void scanForModules(std::map<std::string, Module*> &modules, 
        HTable directories)
{
    HTableIterator iter = table_get_iter(directories);
    if (iter)
    {
        std::string s;
        while (table_iter_next(iter))
        {
            s = std::string(table_iter_get_strs(iter, NULL));
            if (s != "")
            {
                try
                {
                    addModulesFromDir(modules, s);
                }
                catch (MasException &e)
                {
                    masLog.warn(e.getMessage());
                }
                catch (...)
                {
                    masLog.warn("Error loading modules from directory '" + s + "'");
                }
            }
        }
        table_free_iter(iter);
    }
}


class NotDepPred
{
    public:
        bool operator () (std::pair<std::string, Module *> pair)
        {
            return pair.second->depsCount() == 0;
        };
};


static void loadModulesFromMap(std::map<std::string, Module*> &modules)
{
    bool changed = true;
    Module *m;
    std::map<std::string, Module*>::iterator fi;
    
    while ((modules.size() > 0) && changed)
    {
        changed = false;
        fi = find_if(modules.begin(), modules.end(), NotDepPred());
        if (fi != modules.end())
        {
            changed = true;
            m = (*fi).second;
            Loader *loader = Server::getServer()->getLoader(m->getLoader());
            if (! loader)
            {
                masLog.warn("Loader '" + m->getLoader() + "' not found");
            }
            else
            {
                try
                {
                    loader->load(m->getPath(), m->getName(), m->getConf());
                    for (std::map<std::string, Module*>::iterator i = modules.begin();
                            i != modules.end(); i++)
                        (*i).second->removeDep(m->getName());
                }
                catch (MasException &e)
                {
                    masLog.warn(e.getMessage());
                }
                catch (...)
                {
                    masLog.warn("Error loading module '" + m->getName() + "'");
                }
            }
            modules.erase(fi);
        }
    }
}


static void freeModules(std::map<std::string, Module*> &modules)
{
    for (std::map<std::string, Module*>::iterator i = modules.begin();
            i != modules.end(); i++)
        delete (*i).second;
}


void loadModules(HTable directories)
{
    if (! directories)
        return;
    
    std::map<std::string, Module*> modules, remainingModules;
    scanForModules(modules, directories);    
    remainingModules = modules;
    loadModulesFromMap(remainingModules);
    if (remainingModules.size() > 0)
    {
        std::string names;
        for (std::map<std::string, Module*>::iterator i = remainingModules.begin();
                i != remainingModules.end(); i++)
        {
            if (names != "")
                names += ", " + (*i).first;
            else
                names = (*i).first;
        }
        masLog.warn("Some modules not loaded (" + names + ")");
    }
    freeModules(modules);
}

