#include <unistd.h>
#include <signal.h>
#include <string>
#include "server.h"
#include "threadpool.h"
#include "exceptions.h"
#include "log.h"
#include "utils.h"


extern std::string configFile;


static Server *server = NULL;

#ifndef WIN32
static time_t last_signal = 0;


static void sigpipe_handler(int signo)
{
    masLog.warn("SIGPIPE arrived");
}


static void sighup_handler(int signo)
{
    time_t now = time(NULL);
    if (last_signal != now)
    {
        last_signal = now;
        if (server)
        {
            masLog.info("shutting down on signal");
            server->stop();
        }
        else
            masLog.info("server disappeared");
    }
}


static void sigio_handler(int signo)
{
    masLog.warn("SIGIO arrived");
}


static void setSignals()
{
    struct sigaction sa;

    last_signal = 0;
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = sighup_handler;
    sigaction(SIGHUP, &sa, NULL);
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = sighup_handler;
    sigaction(SIGINT, &sa, NULL);
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = sigio_handler;
    sigaction(SIGIO, &sa, NULL);
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = sigpipe_handler;
    sigaction(SIGPIPE, &sa, NULL);
}
#endif


Server::Server(const std::string& configFile): Interruptable()
{
    server = this;
    
    config = table_read(configFile.c_str());
    if (! config)
        throw MasException("Error reading config file '" + configFile + "'");

    masLog.setAgent(table_get_table(config, "log", NULL, NULL));
    
    //threadPool = new ThreadPool(table_get_table(config, "thread_pool", NULL, NULL));
    addLoader("generic", new GenericLoader());

#ifndef WIN32
    setSignals();
#endif

    loadModules();
    
    arena = new Arena(std::string(table_get_strs(config, "streams", "", NULL)));
}


Server::~Server()
{
    if (arena)
        delete arena;

    for (std::map<std::string, RunnablePlugin*>::iterator i = runnablePlugins.begin(); 
            i != runnablePlugins.end(); i++)
        if ((*i).second)
            delete (*i).second;
    
    for (std::map<std::string, Plugin*>::iterator i = plugins.begin(); 
            i != plugins.end(); i++)
    {
        Plugin *p = (*i).second;
        if (p)
            delete p;
    }
    
    for (std::map<std::string, UnitFactory*>::iterator i = unitFactories.begin(); 
            i != unitFactories.end(); i++)
        if ((*i).second)
            delete (*i).second;
    
    for (std::map<std::string, PacketizerFactory*>::iterator i = packetizerFactories.begin(); 
            i != packetizerFactories.end(); i++)
        if ((*i).second)
            delete (*i).second;
    
    for (std::map<std::string, Loader*>::iterator i = loaders.begin(); 
            i != loaders.end(); i++)
        if ((*i).second)
            delete (*i).second;
   
 //   if (threadPool)
 //       delete threadPool;
    if (config)
        table_free(config);
}


void Server::run()
{
    RunnablePlugin *rp;
    
    for (std::map<std::string, RunnablePlugin*>::iterator i = runnablePlugins.begin(); 
            i != runnablePlugins.end(); i++)
    {
        rp = (*i).second;
        if (rp)
            rp->start();
    }
    
    waitForShutdown();
    //sleep(60);
    
    for (std::map<std::string, RunnablePlugin*>::iterator i = runnablePlugins.begin(); 
            i != runnablePlugins.end(); i++)
    {
        if ((*i).second)
            (*i).second->stop();
    }
}


Server* Server::getServer()
{
    return server;
}


/*ThreadPool* Server::getThreadPool()
{
    return threadPool;
}*/


void Server::addPlugin(const std::string& name, Plugin *plugin)
{
    if (plugins.count(name) > 0)
        throw MasException("Plugin '" + name + "' already exists");
    plugins[name] = plugin;
}


void Server::addRunnablePlugin(const std::string& name, RunnablePlugin *plugin)
{
    if (runnablePlugins.count(name) > 0)
        throw MasException("Runnable plugin '" + name + "' already exists");
    runnablePlugins[name] = plugin;
}


Plugin* Server::findPlugin(const std::string& name)
{
    if (plugins.count(name) < 1)
        throw MasException("Plugin '" + name + "' not found");
    return plugins[name];
}


RunnablePlugin* Server::findRunnablePlugin(const std::string& name)
{
    if (runnablePlugins.count(name) < 1)
        throw MasException("Runnable plugin '" + name + "' not found");
    return runnablePlugins[name];
}


HTable Server::getConfig()
{
    return config;
}


void Server::loadModules()
{
    HTable pluginsDirs = table_get_table(config, "modules", NULL, NULL);
    ::loadModules(pluginsDirs);
}

void Server::addLoader(const std::string& name, Loader *loader)
{
    loaders[name] = loader;
}

Loader* Server::getLoader(const std::string& name)
{
    if (loaders.count(name) < 1)
        throw MasException("Loader type '" + name + "' not found");
    return loaders[name];
}

void Server::addUnitFactory(const std::string &name, UnitFactory *factory)
{
    unitFactories[name] = factory;
}

Unit* Server::createUnit(const std::string &name, HTable conf)
{
    if (unitFactories.count(name) < 1)
        throw MasException("Unit type '" + name + "' not found");
    UnitFactory *factory = unitFactories[name];
    if (! factory)
    {
        throw MasException("Invalid unit type '" + name + "'");
    }
    Unit *unit = factory->createUnit(name, conf);
    if (! unit)
        throw MasException("Error creating unit type '" + name + "'");
    return unit;
}


void Server::addPacketizerFactory(const std::string &name, PacketizerFactory *factory)
{
    packetizerFactories[name] = factory;
}

Packetizer* Server::createPacketizer(const std::string &name, PacketizerReceiver *link)
{
    if (packetizerFactories.count(name) < 1)
        throw MasException("Packetizer type '" + name + "' not found");
    PacketizerFactory *factory = packetizerFactories[name];
    if (! factory)
        throw MasException("Invalid packetizer type '" + name + "'");
    Packetizer *packetizer = factory->createPacketizer(name, link);
    if (! packetizer)
        throw MasException("Error creating packetizer type '" + name + "'");
    return packetizer;
}

