#include <sys/time.h>
#include "thread.h"
#include "exceptions.h"
#include "utils.h"


Lock::Lock(bool locked)
{
/*#ifdef WIN32
    InitializeCriticalSection(&mutex);
#else*/
    pthread_mutex_init(&mutex, NULL);
/*#endif*/
    if (locked)
        lock();
}


Lock::~Lock()
{
/*#ifdef WIN32
    DeleteCriticalSection(&mutex);
#else*/
    pthread_mutex_destroy(&mutex);
/*#endif*/
}


void Lock::lock()
{
/*#ifdef WIN32
    EnterCriticalSection(&mutex);
#else*/
    pthread_mutex_lock(&mutex);
/*#endif*/
}


void Lock::unlock()
{
/*#ifdef WIN32
    LeaveCriticalSection(&mutex);
#else*/
    pthread_mutex_unlock(&mutex);
/*#endif*/
}


Cond::Cond()
{
    pthread_cond_init(&cond, NULL);
}

Cond::~Cond()
{
    pthread_cond_destroy(&cond);
}

void Cond::wait(Lock &lock)
{
    pthread_cond_wait(&cond, &lock.mutex);
}

void Cond::wait(Lock &lock, const struct timespec *abstime)
{
    pthread_cond_timedwait(&cond, &lock.mutex, abstime);
}

void Cond::wait(Lock &lock, int sec)
{
    struct timeval tv;
    
    gettimeofday(&tv);
    struct timespec ts;
    ts.tv_sec = tv.tv_sec + sec;
    ts.tv_nsec = tv.tv_usec * 1000;
    pthread_cond_timedwait(&cond, &lock.mutex, &ts);
}

void Cond::signal()
{
    pthread_cond_signal(&cond);
}

void Cond::broadcast()
{
    pthread_cond_broadcast(&cond);
}



Thread::Thread(): Interruptable()
{
    running = false;
}

Thread::~Thread()
{
}


void* runThread(void *data)
{
    Thread *t = (Thread*)data;
    try
    {
        t->running = true;
        t->run();
    }
    catch (...)
    {
    }
    t->running = false;
    if (t->autoDelete)
        delete t;
    return NULL;
}


void Thread::start(bool daemon, bool autoDestroy)
{
    detached = daemon;
    autoDelete = autoDestroy;
    if (pthread_create(&handle, NULL, runThread, (void*)this))
        throw MasException("Error creating thread");
    if (detached)
        pthread_detach(handle);
}

void Thread::stop(bool wait)
{
    pthread_t th = handle;
    Interruptable::stop();
    if (wait && (! detached))
        pthread_join(th, NULL);
}

bool Thread::isRunning()
{
    return running;
}

Locker::Locker()
{
    lock = new Lock(true);
    destroyIt = true;
}

Locker::Locker(Lock& lck)
{
    destroyIt = false;
    lock = &lck;
    lock->lock();
}

Locker::~Locker()
{
    lock->unlock();
    if (destroyIt)
        delete lock;
}



Interruptable::Interruptable()
{
    stopThread = false;
}

Interruptable::~Interruptable()
{
}

void Interruptable::stop()
{
    stopLock.lock();
    stopThread = true;
    stopCond.signal();
    stopLock.unlock();
}

bool Interruptable::isShuttingDown()
{
    stopLock.lock();
    bool b = stopThread;
    stopLock.unlock();
    return b;
}

void Interruptable::waitForShutdown()
{
    stopLock.lock();
    if (stopThread)
    {
        stopLock.unlock();
        return;
    }
    stopCond.wait(stopLock);
    stopLock.unlock();
}


void Interruptable::waitForShutdown(int sec)
{
    stopLock.lock();
    if (stopThread)
    {
        stopLock.unlock();
        return;
    }

    struct timeval tv;
    struct timezone Idunno;
    
    gettimeofday(&tv, &Idunno);
    struct timespec ts;
    ts.tv_sec = tv.tv_sec + sec;
    ts.tv_nsec = tv.tv_usec * 1000;
    stopCond.wait(stopLock, &ts);
    stopLock.unlock();
}


