#include "threadpool.h"
#include "exceptions.h"
#include "log.h"
#include "utils.h"


PoolJob::PoolJob(bool destroy): Thread()
{
    autoDestroy = destroy;
    worker = NULL;
}


PoolJob::~PoolJob()
{
}


bool PoolJob::destroyAfterWorkDone()
{
    return autoDestroy;
}


bool PoolJob::isShuttingDown()
{
    Locker locker(workerLock);
    if (worker)
        return worker->isShuttingDown();
    else
        return false;
}



void PoolJob::run(PoolWorker *worker, void *data)
{
    workerLock.lock();
    this->worker = worker;
    workerLock.unlock();
    try
    {
        run(data);
    }
    catch (...)
    {
    }
    workerLock.lock();
    worker = NULL;
    workerLock.unlock();
}



PoolWorker::PoolWorker(ThreadPool &p, int timeout): pool(p)
{
    job = NULL;
    data = NULL;
    idleTimeout = timeout;
}


PoolWorker::PoolWorker(ThreadPool &p, int timeout, PoolJob *j, void *d): pool(p)
{
    job = j;
    data = d;
    idleTimeout = timeout;
}



PoolWorker::~PoolWorker()
{
}
       
void PoolWorker::run()
{
    jobLock.lock();

    time_t start = time(NULL);
    while (! isShuttingDown())
    {
        if (job)
        {
            job->run(this, data);
            if (job->destroyAfterWorkDone())
                delete job;
            if (isShuttingDown())
                break;
            start = time(NULL);
            job = NULL;
            data = NULL;
            jobLock.unlock();
            pool.markAsIdle(this);
            jobLock.lock();
        }
        else
        {
            if (isShuttingDown())
                break;
            newJobCond.wait(jobLock, idleTimeout);
            if (isShuttingDown())
                break;
            time_t now = time(NULL);
            if ((! job) && (now - start))
                pool.workerTimedOut(this);
        }
    }
    
    jobLock.unlock();
}


void PoolWorker::setJob(PoolJob *j, void *d)
{
    Locker locker(jobLock);
    job = j;
    data = d;
    newJobCond.signal();
}

void PoolWorker::stop(bool waitForTermination)
{
    newJobCond.signal();
    Thread::stop(waitForTermination);
}



ThreadPool::ThreadPool(HTable conf)
{
    maxThreads = table_get_int(conf, "max_threads", 1000, NULL);
    minThreads = table_get_int(conf, "min_threads", 5, NULL);
    idleTimeout = table_get_int(conf, "idle_timeout", 60, NULL);
    maxQueue = table_get_int(conf, "max_queue", 1000, NULL);
    serving = true;
}


ThreadPool::~ThreadPool()
{
    PoolWorker *w;
 
    listsLock.lock();
    serving = false;
    
    for (std::list<PoolWorker*>::iterator i = idleThreads.begin(); 
            i != idleThreads.end(); i++)
    {
        w = (*i);
        w->stop(true);
    }
    
    for (std::list<PoolWorker*>::iterator i = busyThreads.begin(); 
            i != busyThreads.end(); i++)
    {
        w = (*i);
        w->stop(true);
    }
    
    listsLock.unlock();
}


void ThreadPool::addJob(PoolJob *job, void *data, bool force)
{
    Locker locker(listsLock);
    
    if (! serving) return;
    
    if ((! force) && ((int)busyThreads.size() >= maxThreads))
    {
        queueJob(job, data);
        return;
    }

    if (idleThreads.size() > 0)
    {
        PoolWorker *w = (*idleThreads.begin());
        idleThreads.pop_front();
        w->setJob(job, data);
    }
    else
    {
        PoolWorker *w = new PoolWorker(*this, idleTimeout, job, data);
        busyThreads.push_back(w);
        w->start(false, true);
    }
}


void ThreadPool::queueJob(PoolJob *job, void *data)
{
    if ((int)queue.size() >= maxQueue)
        throw MasException("Job queue full");
    queue.push_back(Job(job, data));
}


void ThreadPool::markAsIdle(PoolWorker *worker)
{
    Locker locker(listsLock);

    if ((int)busyThreads.size() > maxThreads)
    {
        busyThreads.remove(worker);
        worker->stop(false);
        return;
    }
    
    if (queue.size() > 0)
    {
        Job job = *queue.begin();
        queue.pop_front();
        worker->setJob(job.job, job.data);
    }
    else
    {
        busyThreads.remove(worker);
        idleThreads.push_back(worker);
    }
}


void ThreadPool::workerTimedOut(PoolWorker *worker)
{
    Locker locker(listsLock);

    if ((int)idleThreads.size() + (int)busyThreads.size() > minThreads)
    {
        idleThreads.remove(worker);
        worker->stop(false);
    }
}


