#ifndef WIN32
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log.h"
#include "server.h"
#include "exceptions.h"
#include "guardian.h"


#define DROP_INTERVAL  10
#define DROP_COUNT     5


static time_t last_drop = 0;
static int drop_count = 0;


extern std::string configFile;


static int restart_worker()
{
    if (fork() == 0)
    {
        try
        {
            Server server(configFile);
            server.run();
        }
        catch (MasException& e)
        {
            masLog.error(e.getMessage());
        }
        catch (...)
        {
            masLog.error("worker exception");
        }
        exit(0);
    }
    
    return 0;
}


static void sigchild_handler(int signo)
{
    time_t now;
    int status;
    
//    log_message(LOG_WARN, "SIGCHLD arrived");

    wait(&status);

    if (WIFEXITED(status) && (WEXITSTATUS(status) == 2))
        exit(0);
    
    now = time(NULL);
    if (now - last_drop > DROP_INTERVAL)
        drop_count = 1;
    else
        drop_count++;
    last_drop = now;

    if (drop_count > DROP_COUNT)
    {
        masLog.error("respawning too fast.  waiting a minute");
        usleep(60000000);
    }

    usleep(1000000);
    restart_worker();
}


int start_guardian()
{
    pid_t pid;
    struct sigaction sa;

    pid = fork();
    if (pid < 0)
        masLog.error("can't fork");
    if (! pid)
        return 0;

    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = sigchild_handler;
    sigaction(SIGCHLD, &sa, NULL);
    
    while (1)
        usleep(1000000);
    
    return 0;
}

#else

#include <windows.h>
#include "guardian.h"
#include "server.h"
#include "log.h"
#include "utils.h"


static SERVICE_STATUS_HANDLE status_hndl;
static SERVICE_STATUS status;


Server *server;


void WINAPI service_handler(DWORD fdwControl)
{
    switch (fdwControl)
    {
	case SERVICE_CONTROL_STOP:
	    status.dwCurrentState = SERVICE_STOP_PENDING;
	    SetServiceStatus(status_hndl, &status);
            server->stop();
	   /* if (stop_worker())
	    {
		log_message(LOG_ERROR, "Termination error in service_handler");
		status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
		status.dwServiceSpecificExitCode = 2;
	    }
	    else
		log_message(LOG_MSG, "Successfully terminated service_handler");*/
	    status.dwCurrentState = SERVICE_STOPPED;
	    break;
	case SERVICE_CONTROL_INTERROGATE:
	    break;
	default:
	    masLog.warn("Unknown command " + numToStr((int)fdwControl));
    }
    
    if (! SetServiceStatus(status_hndl, &status))
	masLog.error("Error setting status...");
}


extern std::string configFile;


extern "C" {
void WINAPI maserver_service_main(DWORD argc, LPTSTR *argv);
}

void WINAPI maserver_service_main(DWORD argc, LPTSTR *argv)
{
    masLog.info("Starting main service");
    status_hndl = RegisterServiceCtrlHandler("maserver", service_handler);
    if (! status_hndl)
    {
	masLog.error("Error registering service handler");
	return;
    }

    //LogString("initialization started");
    status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    status.dwCurrentState = SERVICE_START_PENDING;
    status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    status.dwWin32ExitCode = NO_ERROR;
    status.dwCheckPoint = 0;
    status.dwWaitHint = 0;
    SetServiceStatus(status_hndl, &status);

    try
    {
        server = new Server(configFile);
    }
    catch (...)
    {
	status.dwCurrentState = SERVICE_STOP_PENDING;
	SetServiceStatus(status_hndl, &status);
	masLog.error("Error initializing server");
	status.dwCurrentState = SERVICE_STOPPED;
	status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
	status.dwServiceSpecificExitCode = 1;
	SetServiceStatus(status_hndl, &status);
	return;
    }
    status.dwCurrentState = SERVICE_RUNNING;
    SetServiceStatus(status_hndl, &status);
    //LogString("initialization finished");

    try
    {
        server->run();
    }
    catch (...)
    {
    }

    status.dwCurrentState = SERVICE_STOP_PENDING;
    SetServiceStatus(status_hndl, &status);

    delete server;
  /*  if (pua_done())
    {
	LogString("Termination error in service_main");
	status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
	status.dwServiceSpecificExitCode = 2;
    }
    else
	LogString("Successfully terminated service_main");*/
    status.dwCurrentState = SERVICE_STOPPED;
    SetServiceStatus(status_hndl, &status);
}



int start_service()
{
    SERVICE_TABLE_ENTRY ServiceStartTable[] = {
	{ "maserver", maserver_service_main },
	{ NULL, NULL }
    };
   
   // LogString("Service started");
    if (! StartServiceCtrlDispatcher(ServiceStartTable))
    {
	masLog.error("Start service error");
	return 1;
    }
   // LogString("Service successfull shutdown");
    return 0;
}

#endif

