#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/resource.h>
#include <signal.h>
#include <unistd.h>
#else
#include <winsock2.h>
#endif
#include <iostream>

#include "conf.h"
#include "log.h"
#include "guardian.h"
#include "server.h"
#include "exceptions.h"


static int nodaemon = 0;


std::string configFile;


static void print_help(int terminate)
{
    printf("USAGE:\n");
    printf("  maserver [options]\n");
#ifdef WIN32
    printf("OR\n");
    printf("  maserver --install-service\n");
#endif
    printf("OPTIONS:\n");
    printf("  --nodaemon      run in foreground\n");
    printf("  --conf <file>   config filename\n");
    printf("  --version       show version number\n");
    printf("  --help          this help screen\n");
    if (terminate >= 0)
        exit(terminate);
}


static void print_version(int terminate)
{
    printf("Multicast Radio Server 0.8.1\n");
    printf("Copyright (c) 2002-2003 Alexander Babichev\n");
    if (terminate >= 0)
        exit(terminate);
}


#ifdef WIN32
static void pferror(char *msg, ...)
{
    va_list list;
    va_start(list, msg);
    vfprintf(stderr, msg, list);
    va_end(list);
    exit(1);
}


/*static std::string extractPrefix(char *programPrefix)
{
    int length;
    int slashCount = 0;
    
    if (! programPrefix)
        return "";

    length = strlen(programPrefix);
    int i = length - 1;
    while (i > 0)
    {
        if ((programPrefix[i] == '\\') || (programPrefix[i] == '/'))
            slashCount++;
        if (slashCount == 2)
            return std::string(programPrefix, i);
        i--;
    }

    return "";
}*/


static void install_service(const std::string prefix)
{
    SC_HANDLE sch, svh;

    sch = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (! sch)
    {
	pferror("Error connecting to manager.\n\r");
        exit(1);
    }

    std::string exeName(prefix + "\\bin\\maserver.exe --conf \"" + prefix + "\\etc\\maserver.conf\"");
    
    svh = CreateService(sch, "maserver", "maserver", SERVICE_ALL_ACCESS, 
	    SERVICE_WIN32_OWN_PROCESS, SERVICE_AUTO_START, SERVICE_ERROR_NORMAL,
	    exeName.c_str(), NULL, NULL, NULL, NULL, NULL);
    if (! svh)
	pferror("Error creating service %i.\n\r", GetLastError());
    else
    {
        SERVICE_DESCRIPTION sd;
        sd.lpDescription = "Multicast audio server";
        ChangeServiceConfig2(svh, SERVICE_CONFIG_DESCRIPTION, &sd);
    }
    CloseServiceHandle(svh);
    CloseServiceHandle(sch);
   
    std::cout << "Successfully installed." << std::endl;
    exit(0);
}


static void remove_service()
{
    SC_HANDLE sch, svh;
    int res;

    sch = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (! sch)
    {
	pferror("Error connecting to manager.\n\r");
        exit(1);
    }

    svh = OpenService(sch, "maserver", SERVICE_ALL_ACCESS);
    res =DeleteService(svh);
    if (res)
	pferror("Error deleting service.\n\r");
    else
        std::cout << "Successfully removed." << std::endl;

    CloseServiceHandle(svh);    
    CloseServiceHandle(sch);

    exit(0);
}
#endif


static void parse_args(int argc, char *argv[])
{
    int i;
    
    if (argc == 1)
        return;

    for (i = 1; i < argc; i++)
    {
        if (! argv[i])
            continue;
        if ((! strcmp(argv[i], "--conf")) && (i < argc - 1))
            configFile = std::string(argv[++i]);
        else if (! strcmp(argv[i], "--nodaemon"))
            nodaemon = 1;
        else if (! strcmp(argv[i], "--help"))
            print_help(0);
        else if (! strcmp(argv[i], "--version"))
            print_version(0);
#ifdef WIN32
        else if ((! strcmp(argv[i], "--install-service")) && (i < argc - 1))
            install_service(argv[++i]);
        else if (! strcmp(argv[i], "--remove-service"))
            remove_service();
#endif
        else
        {
            printf("Invalid option '%s'.\n", argv[i]);
            print_help(1);
        }
    }
}


#ifndef WIN32
static void demonize()
{
    unsigned int fd;
    struct rlimit flim;

    if (getppid() != 1)
    {
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, SIG_IGN);

	if (fork()) 
        {
            exit(0);
        }

	setsid();
    }

    getrlimit(RLIMIT_NOFILE, &flim);
    for (fd=0; fd < flim.rlim_max; fd++)
	close(fd); 

    chdir("/tmp");
}
#endif


static void read_server_config()
{
    HTable table;

    table = table_read(configFile.c_str());
    if (! table)
        throw MasException(std::string("Error reading config file '") + configFile + "'");
    table_free(table);
}

   
#ifdef WIN32
static void initWinsock()
{
    WSADATA wsadata;
    
    if (WSAStartup(MAKEWORD(2, 0), &wsadata))
        throw MasException("Error initializing winsock 2");
}
#endif


int main(int argc, char *argv[])
{
    configFile = CONFPATH;

    parse_args(argc, argv);

#ifdef WIN32
    if (! nodaemon)
    {
        initWinsock();
        return start_service();
    }
#endif
    
    int seed = time(NULL);
    srand(seed);
#ifndef WIN32
    srandom(seed);
    srand48(seed);
#endif

    try 
    {
        read_server_config();

#ifndef WIN32
        if (! nodaemon)
        {
            demonize();
            if (start_guardian())
                return 0;
        }
       
        Server *server = new Server(configFile);
        server->run();
        delete server;
#else
        initWinsock();
        
        Server server(configFile);
        server.run();
#endif

    }
    catch (MasException& e)
    {
        masLog.error(e.getMessage());
    }
    catch (...)
    {
        masLog.error("unknown server exception");
    }

#ifdef WIN32
    if (nodaemon)
        WSACleanup();
#endif
    
    return 0;
}

