#include "decode.h"



class DECODEIMPORT MediaDecoder: public SingleInputOutputUnit
{
    private:
        Decoder *decoder;
        MediaType *mediaType;
        DecoderRegistry *registry;

        class Preference
        {
            public:
                std::string typeName;
                std::string decName;
                HTable conf;

            public:
                Preference() { };
                Preference(const Preference &p);
                Preference(HTable conf);
                ~Preference() { };
        };
        
        std::map<std::string, Preference> preferences;
    
    public:
        MediaDecoder();
        MediaDecoder(HTable conf);
        virtual ~MediaDecoder();
        
    public:
        virtual std::set<MediaType> getAcceptedMediaTypes();
        virtual std::string getType() { return "decoder"; };
        
    public:
        virtual void run();

    private:
        Decoder *createDecoder(MediaType &mediaType);
        void readPreferences(HTable conf);
};


MediaDecoder::Preference::Preference(const Preference &p)
{
    typeName = p.typeName;
    decName = p.decName;
    conf = p.conf;
}

MediaDecoder::Preference::Preference(HTable table)
{
    typeName = table_get_strs(table, "stream", "", NULL);
    decName = table_get_strs(table, "decoder", "", NULL);
    conf = table;
}


MediaDecoder::MediaDecoder(): SingleInputOutputUnit()
{
    decoder = NULL;
    mediaType = NULL;
    registry = (DecoderRegistry*)Server::getServer()->findPlugin("decoderRegistry");
    if (! registry)
        throw MasException("decoderRegistry plugin not found");
}


MediaDecoder::MediaDecoder(HTable conf): SingleInputOutputUnit(conf)
{
    decoder = NULL;
    mediaType = NULL;
    registry = (DecoderRegistry*)Server::getServer()->findPlugin("decoderRegistry");
    if (! registry)
        throw MasException("decoderRegistry plugin not found");
    readPreferences(table_get_table(conf, "preferred", NULL, NULL));
}


MediaDecoder::~MediaDecoder()
{
    if (decoder)
        delete decoder;
    if (mediaType)
        delete mediaType;
}


Decoder* MediaDecoder::createDecoder(MediaType &mediaType)
{
    std::string s(mediaType.getType() + "/" + mediaType.getSubtype());
    if (preferences.count(s) > 0)
    {
        Preference &p = preferences[s];
        return registry->createDecoder(mediaType, output, p.conf);
    }
    else
        return registry->createDecoder(mediaType, output, NULL);
}


void MediaDecoder::readPreferences(HTable conf)
{
    if (! conf)
        return;
    HTableIterator iter = table_get_iter(conf);
    if (iter)
    {
        HTable prefTable;
        while (table_iter_next(iter))
        {
            prefTable = table_iter_get_table(iter, NULL);
            if (prefTable)
            {
                Preference p(prefTable);
                preferences[p.typeName] = p;
            }
            else
                masLog.error("Invalid entry in preferred decoders table");
        }
        table_free_iter(iter);
    }
}


void MediaDecoder::run()
{
    MediaType *mtype, *mediaType=NULL;
    MediaInfo *minfo;
    unsigned char *buffer;
    int bufSize = 1024, length;
    unsigned long long int pos = 0L;
    double playTime;
   
    buffer = new unsigned char[bufSize];
    
    while (! isShuttingDown())
    {
        if (input && output)
        {
            input->beginGet();
            try
            {
                input->getPacketInfo(mtype, minfo, length, playTime, pos);
                if ((! decoder) || (! mediaType) || (! mtype->isEqual(*mediaType)))
                {
                    if (mediaType) { delete mediaType; mediaType = NULL; }
                    if (decoder) { delete decoder; decoder = NULL; }
                    mediaType = mtype->clone();
                    decoder = createDecoder(*mediaType);
                }
                if (bufSize < length)
                {
                    delete[] buffer;
                    bufSize = length;
                    buffer = new unsigned char[bufSize];
                }
                input->getData(pos, buffer);
            }
            catch (MasException &e)
            {
                masLog.error(e.getMessage());
            }
            catch (...)
            {
                masLog.error("Error decoding stream");
            }
            if (decoder)
            {
                try
                {
                    decoder->addData(buffer, length, playTime, *mediaType);
                }
                catch (MasException &e)
                {
                    masLog.error(e.getMessage());
                }
                catch (...)
                {
                    masLog.error("Error decoding stream buffer");
                }
            }
            input->endGet();
        }
        else
            sleep(1);
    }
}


std::set<MediaType> MediaDecoder::getAcceptedMediaTypes()
{
    std::set<MediaType> sets;
    sets.insert(MediaType("*", "*"));
    return sets;
}



DecoderRegistry::~DecoderRegistry()
{
    for (std::map<std::string, DecoderFactory*>::iterator i = factories.begin();
            i != factories.end(); i++)
        if ((*i).second)
            delete (*i).second;
}

void DecoderRegistry::registerDecoder(const std::string &name, DecoderFactory *factory)
{
    if (! factory)
        throw MasException("Invalid decoder factory to register");
    factories[name] = factory;
}

Decoder* DecoderRegistry::createDecoder(const MediaType &mediaType, 
        PacketizerReceiver *receiver, HTable preferred)
{
    DecoderFactory *f;
    
    if (preferred)
    {
        std::string name(table_get_strs(preferred, "decoder", "", NULL));
        if ((name != "") && (factories.count(name) > 0))
        {
            f = factories[name];
            if (f && f->isSupportedMedia(mediaType))
                return f->createDecoder(mediaType, receiver, preferred);
        }
    }
    
    for (std::map<std::string, DecoderFactory*>::iterator i = factories.begin();
            i != factories.end(); i++)
    {
        f = (*i).second;
        if (f && f->isSupportedMedia(mediaType))
            return f->createDecoder(mediaType, receiver, NULL);
    }
    throw MasException("No suitable decoder for '" + mediaType.getType() + "/" +
            mediaType.getSubtype() + "' media type");
}


class MediaDecoderFactory: public UnitFactory
{
    public:
        MediaDecoderFactory() { };
        virtual ~MediaDecoderFactory() { };
        virtual Unit* createUnit(const std::string &name, HTable conf);
};

Unit* MediaDecoderFactory::createUnit(const std::string &name, HTable conf)
{
    if (name != "decoder")
        throw MasException("Invalid decoder unit type'" + name + "'");
    if (conf)
        return new MediaDecoder(conf);
    else
        return new MediaDecoder();
}


extern "C" int EXPORT_REGFUNC register_decode_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addPlugin(std::string("decoderRegistry"), 
                new DecoderRegistry());
        Server::getServer()->addUnitFactory(std::string("decoder"), 
                new MediaDecoderFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

