#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include "fileinput.h"
#include "fileformats.h"


FileInput::FileInput(): OutputUnit()
{
    playlistHandler = NULL;
}


FileInput::FileInput(HTable config): OutputUnit(config)
{
    HTable playlistTable;
    std::string handlerType;

    if (! config)
        throw MasException("Files input config missing");
    playlistHandler = NULL;
    playlistTable = table_get_table(config, "playlist", NULL, NULL);
    if (! playlistTable)
        throw MasException("Playlist parameters missing");
    handlerType = std::string(table_get_strs(playlistTable, "type", "", NULL));
    playlistHandler = createPlaylistHandler(handlerType, playlistTable);
}


FileInput::~FileInput()
{
    if (playlistHandler)
        delete playlistHandler;
}


std::set<MediaType> FileInput::getAcceptedMediaTypes()
{
    std::set<MediaType> s;
    s.insert(MediaType("*", "*"));
    return s;
}


void FileInput::run()
{
    struct timeval tv;
    gettimeofday(&tv);
    startTime = tv.tv_sec * 10000.0 + tv.tv_usec / 100.0;
    playingTime = 0.0;
    
    if (! playlistHandler)
        throw MasException("Playlist handler missing");
    
    FileFormatRegistry *formatsRegistry;
    formatsRegistry = (FileFormatRegistry*)Server::getServer()->findPlugin("fileformats");
    if (! formatsRegistry)
        throw MasException("Can't find file formats registry");

    int attempts = 0;
    while (! isShuttingDown())
    {
        try
        {
            std::string filename(playlistHandler->getNextSong());
            masLog.debug("playing " + filename);

            if (filename != "")
            {
                FileFormat *format = formatsRegistry->createFileFormat(filename);
                if (! format)
                    throw MasException("Unknown format of '" + filename + "'");
                bool deleteFormat = true;
                try
                {
                    if (output)
                    {
                        MediaInfo minfo;
                        if (format->hasMediaInfo())
                            minfo = format->getMediaInfo();
                        patchMediaInfo(minfo);
                        output->setMediaInfo(minfo);
                    }
                    format->packetize(this, this);
                    deleteFormat = false;
                    delete format;
                    attempts = 0;
                }
                catch (MasException &e)
                {
                    if (deleteFormat) delete format;
                    throw e;
                }
                catch (...)
                {
                    if (deleteFormat) delete format;
                    throw MasException("Unknown file format exception in fileinput");
                }
            }
            else
                attempts++;
        }
        catch (MasException &e)
        {
            attempts++;
            masLog.error(e.getMessage());
        }
        catch (...)
        {
            attempts++;
            masLog.error("Unknown exception in file input");
        }

        if (attempts > 5)
        {
            sleep(1);
            attempts = 0;
        }
    }
}


void FileInput::addData(unsigned char *data, int length, double playTime, 
        const MediaType &mediaType)
{
    struct timeval tv;

    if (output)
        output->addData(data, length, playTime, mediaType);

    gettimeofday(&tv);
    playingTime = playingTime + playTime / 100;
    double realPlayTime = (tv.tv_sec * 10000.0 + tv.tv_usec / 100.0) -
        startTime;
    long long int pause = (long long int)((playingTime - realPlayTime) * 100.0);
    //pause -= 1000000;
    if (pause > 0)
        usleep(pause);
}


Unit* FileUnitFactory::createUnit(const std::string &name, HTable conf)
{
    if (name != "fileinput")
        throw MasException("Invalid file input unit type'" + name + "'");
    if (conf)
        return new FileInput(conf);
    else
        return new FileInput();
}


extern "C" int EXPORT_REGFUNC register_fileinput_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addUnitFactory(std::string("fileinput"), 
                new FileUnitFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

