#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <iostream>
#include <algorithm>
#include <fstream>
#include "intpllist.h"
#include "exceptions.h"
#include "log.h"
#include "random.h"
#include "utils.h"


typedef std::vector<std::string>::iterator strIterator;


InternalPlaylistHandler::InternalPlaylistHandler()
{
  //  type = "internal";
    current_song = -1;
    order = Shuffle;
    updatePlaylist = 1;
}


InternalPlaylistHandler::InternalPlaylistHandler(HTable table)
{
    std::string orderName;
    
    if (! table)
        throw MasException("Missing internal playlist parameters");
    
  //  type = "internal";
    current_song = -1;

    playlistFilename = std::string(table_get_strs(table, "playlist", "", NULL));
    blockFilename = std::string(table_get_strs(table, "blockfile", "", NULL));
    operDir = std::string(table_get_strs(table, "subplaylists", "", NULL));
    blockOper = std::string(table_get_strs(table, "blocksub", "", NULL));
    orderName = std::string(table_get_strs(table, "order", "", NULL));
    if (orderName == "forward")
        order = Forward;
    else if (orderName == "backward")
        order = Backward;
    else if (orderName == "random")
        order = Random;
    else if (orderName == "shuffle")
        order = Shuffle;
    else
        throw MasException("Invalid order type '" + orderName + "'");
    updatePlaylist = (table_get_int(table, "reload", 0, NULL) != 0);

    readPlaylist();
}


InternalPlaylistHandler::~InternalPlaylistHandler()
{
    if (deleteFile != "")
    {
        unlink(std::string(operDir + "/" + deleteFile).c_str());
        deleteFile = "";
    }
}


static std::string getFileExtLower(char *filename)
{
    if (! filename)
        return "";
    
    int len = strlen(filename);

    if (len > 0)
    {
        int i = len - 1;
        while (i)
        {
            if (filename[i] == '.')
            {
                if (i == len - 1)
                    break;
                i++;
                char *buf = strdup(filename + i);                
                for (char *c = buf; *c; c++)
                    *c = tolower(*c);
                std::string s(buf);
                free(buf);
                return s;
            }
            i--;
        }
    }
    return "";
}


static bool fileExists(const std::string& filename)
{
    std::ifstream ifs(filename.c_str());
    int res = ifs.good();
        ifs.close();
    return res ? true: false;
}


void InternalPlaylistHandler::readSubPlaylist(const std::string& filename)
{
    try
    {
        std::ifstream ifs(filename.c_str());
        if (! ifs.good())
            throw MasException("Error opening playlist '" + filename + "'");
        std::string s;
        while (std::getline(ifs, s))
        {
            if (s != "")
                operList.push_back(s);
        }
        ifs.close();
    }
    catch (MasException& e)
    {
        throw e;
    }
    catch (...)
    {
        throw MasException("Error reading playlist '" + filename + "'");
    }
}


std::string InternalPlaylistHandler::getNextOperSong()
{
    if (blockOper != "")
        if (fileExists(blockOper))
            return "";
    
    DIR *dir;
#ifndef WIN32
    struct dirent de, *nde;
#else
    struct dirent *nde;
#define de (*nde)
#endif

    dir = opendir(operDir.c_str());
    if (! dir)
        return "";

#ifdef WIN32
    while (! (nde = readdir(dir)))
#else
#ifdef sun
    while (! (nde = readdir_r(dir, &de)))
#else
    while (! readdir_r(dir, &de, &nde))
#endif
#endif
    {
        if (! nde)
            break;
        if (de.d_name[0] != '.')
        {
            std::string ext = getFileExtLower(de.d_name);
            if (ext == "mp3")
            {
                closedir(dir);
                std::string filename = operDir + "/" + de.d_name;
                deleteFile = filename;
                return filename;
            }
            else
            {
                if ((ext == "m3u") || (ext == "txt"))
                {
                    std::string filename = operDir + "/" + de.d_name;
                    try
                    {
                        readSubPlaylist(filename);
                    }
                    catch (MasException& e)
                    {
                        masLog.warn(e.getMessage());
                    }
                    catch (...)
                    {
                    }
                    unlink(filename.c_str());
                }
            }
        }
    }

    closedir(dir);
    return "";
}


std::string InternalPlaylistHandler::getNextSong()
{
    std::string operSong("");
    
    if (deleteFile != "")
    {
        unlink(deleteFile.c_str());
        deleteFile = "";
    }
    
    if (operDir != "")
    {
        operSong = getNextOperSong();
        if (operSong != "")
            return operSong;
    }
    
    if (operList.size() > 0)
    {
        std::string curSong = operList[0];
        operList.erase(operList.begin());
        return curSong;
    }
    
    if (updatePlaylist && isPlaylistUpdated())
        readPlaylist();
    
    if (songs.size() < 1)
        return "";

    switch (order)
    {
        case Shuffle:
        case Forward:
            current_song++;
            if (current_song >= (int)songs.size())
                current_song = 0;
            break;
            
        case Backward:
            current_song--;
            if (current_song < 0)
                current_song = songs.size() - 1;
            break;

        default:
            current_song = (int) ((double)(songs.size() - 1) * rndGen.genReal1());
            if ((current_song < 0) || (current_song >= (int)songs.size()))
                throw MasException("Random generator error");
    }

    return songs[current_song];
}


bool InternalPlaylistHandler::isPlaylistUpdated()
{
    struct stat st;
    
    if (blockFilename != "")
        if (fileExists(blockFilename))
            return false;
    
    if (stat(playlistFilename.c_str(), &st))
        throw MasException("Can't get information about playlist '" + 
                playlistFilename + "'");

    if ((lastPlaylistSize != st.st_size) || (lastPlaylistCtime != st.st_ctime) 
            || (lastPlaylistMtime != st.st_mtime))
        return true;
    else
        return false;
}


void InternalPlaylistHandler::readPlaylist()
{
    struct stat st;
    
    if (stat(playlistFilename.c_str(), &st))
        throw MasException("Can't get information about playlist '" + 
                playlistFilename + "'");

    lastPlaylistSize = st.st_size;
    lastPlaylistMtime = st.st_mtime;
    lastPlaylistCtime = st.st_ctime;
    
    try
    {
        std::ifstream ifs(playlistFilename.c_str());
        if (! ifs.good())
            throw MasException("Error opening playlist '" + playlistFilename + "'");
        songs.clear();
        std::string s;
        while (std::getline(ifs, s))
        {
            if (s != "")
                songs.push_back(s);
        }
        ifs.close();
        if (order == Shuffle)
        {
            std::random_shuffle<strIterator>(songs.begin(), songs.end());
            std::random_shuffle<strIterator>(songs.begin(), songs.end());
            std::random_shuffle<strIterator>(songs.begin(), songs.end());
        }
    }
    catch (MasException& e)
    {
        throw e;
    }
    catch (...)
    {
        throw MasException("Error reading playlist '" + playlistFilename + "'");
    }
}


PlaylistHandler* InternalPlaylistFactory::createPlaylist(HTable args)
{
    if (args)
        return new InternalPlaylistHandler(args);
    else
        return new InternalPlaylistHandler();
}

InternalPlaylistFactory::InternalPlaylistFactory()
{
}

InternalPlaylistFactory::~InternalPlaylistFactory()
{ 
}
        

extern "C" int EXPORT_REGFUNC register_intpllist_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addPlugin("playlist::internal", 
                new InternalPlaylistFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

