#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <pthread.h>
#ifdef WIN32
#include <winsock.h>
#endif
#include "httpoutput.h"
#include "fileformats.h"


#define META_DATA_INTERVAL 204800



HttpPeer::HttpPeer(Link *alink, HttpdConnection *conn, HttpOutput *out,
        const std::string& onConnect)
{
    link = alink;
    connection = conn;
    output = out;
    useIcyMetadata = false;
    introFilename = onConnect;
}


HttpPeer::~HttpPeer()
{
}


void HttpPeer::sendIcyMeta()
{
    std::string title;

 /*   if (minfo.hasProperty("name"))
        title = minfo.getProperty("name");
    else
    { */
        if (minfo.hasProperty("title") && minfo.hasProperty("artist"))
            title = minfo.getProperty("artist") + " - " + minfo.getProperty("title");
        else
            if (minfo.hasProperty("title"))
                title = minfo.getProperty("title");
            else
                if (minfo.hasProperty("artist"))
                    title = minfo.getProperty("artist");
                else
                    if (minfo.hasProperty("name"))
                        title = minfo.getProperty("name");
   /* } */
    
    std::string url = minfo.getProperty("url");
    int len = title.length();
    for (int i = 0; i < len; i++)
    {
        if (title[i] == '\'')
            title[i] = '"';
        else if (title[i] == '\\')
            title[i] = '/';
    }
    title = "StreamTitle='" + title + "';StreamUrl='" + url + "';";
    len = title.length() + 1;
    int td = len / 16;
    int alignedLen;
    while (((alignedLen = (td * 16))) < len)
        td++;
    if (td > 250)
    {
        title = "StreamTitle='Too long'";
        len = title.length();
        td = 2;
        alignedLen = 32;
    }
    unsigned char cb = (unsigned char)td;
    connection->sendData(&cb, 1);
    connection->sendData(title);
    cb = '\0';
    for (int i = len - 1; i < alignedLen; i++)
        connection->sendData(&cb, 1);
}


#define BUF_SIZE 2048


void HttpPeer::sendData(unsigned char *data, int length)
{
    if (useIcyMetadata)
    {
        if (bytesLeftBeforeMeta < length)
        {
            if (bytesLeftBeforeMeta > 0)
                connection->sendData(data, bytesLeftBeforeMeta);
            sendIcyMeta();
            if (length - bytesLeftBeforeMeta > 0)
                connection->sendData(data + bytesLeftBeforeMeta, 
                        length - bytesLeftBeforeMeta);
            bytesLeftBeforeMeta = META_DATA_INTERVAL - (length - bytesLeftBeforeMeta);
        }
        else
        {
            connection->sendData(data, length);
            bytesLeftBeforeMeta -= length;
        }
    }
    else
        connection->sendData(data, length);
}


void HttpPeer::run()
{
    unsigned char *buf;
    int bufSize = 1024;
    unsigned long long int pos = 0;
    int readed;
    MediaType *mediaType;
    MediaInfo *mediaInfo;
    double playTime;
    
    sendHeader();
   
    if (introFilename != "")
        playIntro();
    
    buf = (unsigned char*)malloc(bufSize);
    if (! buf)
        throw MasException("Error allocating http peer buffer");
    
    while (! connection->isShuttingDown())
    {
        link->beginGet();
        link->getPacketInfo(mediaType, mediaInfo, readed, playTime, pos);
        if (readed > 0)
        {
            if (! (*mediaInfo == minfo))
            {
                minfo = *mediaInfo;
            }
            if (readed > bufSize)
            {
                bufSize = readed + 10;
                unsigned char *d = (unsigned char*)realloc(buf, bufSize);
                if (! d)
                {
                    link->endGet();
                    free(buf);
                    throw MasException("Error growing http peer buffer");
                }
            }
            link->getData(pos, buf);
        }
        link->endGet();
        if (readed > 0)
            sendData(buf, readed);
    }
}


void HttpPeer::sendHeader()
{
    std::string mimeType;
    unsigned long long int pos = 0;
    MediaType *mediaType;
    MediaInfo *mediaInfo;
    int size;
    double playTime;
    
    link->beginGet();
    link->getPacketInfo(mediaType, mediaInfo, size, playTime, pos);
    minfo = *mediaInfo;
    mimeType = mediaType->getMimeType();
    link->endGet();
    
    std::string icyMeta = connection->getHeader("Icy-MetaData");
    if (icyMeta == "1")
    {
        bytesLeftBeforeMeta = META_DATA_INTERVAL;
        useIcyMetadata = true;
    }
 
    connection->setStatus(200);
    connection->setContentType(mimeType);
    connection->setHttpHeader("Cache-Control", "no-cache");
    connection->setHttpHeader("Pragma", "no-cache");
    if (minfo.hasProperty("name"))
        connection->setHttpHeader("icy-name", minfo.getProperty("name"));
    if (minfo.hasProperty("genre"))
        connection->setHttpHeader("icy-genre", minfo.getProperty("genre"));
    if (minfo.hasProperty("url"))
        connection->setHttpHeader("icy-url", minfo.getProperty("url"));
    if (useIcyMetadata)
        connection->setHttpHeader("icy-metaint", numToStr(META_DATA_INTERVAL));
    //connection->setContentLength(540000000);
    connection->transferHeaders();
}


void HttpPeer::playIntro()
{
    gettimeofday(&curTime);
    try
    {
        FileFormatRegistry *formatsRegistry = (FileFormatRegistry*)
            Server::getServer()->findPlugin("fileformats");
        if (! formatsRegistry)
            throw MasException("Can't find file formats registry");        
        FileFormat *format = formatsRegistry->createFileFormat(introFilename);
        if (! format)
            throw MasException("Unknown format of intro '" + introFilename + "'");
        bool deleteFormat = true;
        try
        {
          /*  if (format->hasMediaInfo() && output)
            {
                MediaInfo minfo = format->getMediaInfo();
                output->setMediaInfo(minfo);
            } */
            format->packetize(connection, this);
            deleteFormat = false;
            delete format;
        }
        catch (MasException &e)
        {
            if (deleteFormat) delete format;
            throw e;
        }
        catch (...)
        {
            if (deleteFormat) delete format;
            throw MasException("Unknown file format exception in fileinput");
        }
        
        /*MP3Packetizer packetizer(*this);
        SongInfo songInfo;
        getSongInfo(introFilename, songInfo);
    
        gettimeofday(&curTime, &Idunno);
        packetizeMP3File(packetizer, introFilename, songInfo.start, songInfo.size);*/
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
    }
    catch (...)
    {
        masLog.error("Error playing http intro");
    }
}


void HttpPeer::addData(unsigned char *data, int length, double playTime,
                const MediaType &mediaType)
{
    struct timeval presentTime;
    
    curTime.tv_usec += (int)playTime;
    while (curTime.tv_usec >= 1000000)
    {
        curTime.tv_sec++;
        curTime.tv_usec -= 1000000;
    }

    sendData(data, length);

    long long int shouldBe, now;
    shouldBe = curTime.tv_sec * 1000000 + curTime.tv_usec; 
    gettimeofday(&presentTime);
    now = presentTime.tv_sec * 1000000 + presentTime.tv_usec; 
    long long int pause = shouldBe - now;
    if (pause > 0)
        usleep(pause);
}


HttpOutput::HttpOutput()
{
    httpd = (Httpd*)Server::getServer()->findRunnablePlugin("httpd");
    if (! httpd)
        throw MasException("Httpd plugin not running");
}


HttpOutput::HttpOutput(HTable params): SingleInputUnit(params)
{
    mount = std::string(table_get_strs(params, "mount", "", NULL));
    if (mount == "")
        throw MasException("Invalid http mount point");
    onConnect = std::string(table_get_strs(params, "intro", "", NULL));
    httpd = (Httpd*)Server::getServer()->findRunnablePlugin("httpd");
    if (! httpd)
        throw MasException("Httpd plugin not running");
    httpd->addHandler(mount, this);
}


HttpOutput::~HttpOutput()
{
    if (mount != "")
        httpd->removeHandler(mount);
}


void HttpOutput::run()
{    
    while (! isShuttingDown())
        waitForShutdown();
}


int HttpOutput::handleRequest(HttpdConnection& connection, const ServerURL& url, 
                enum RequestType requestType)
{
    HttpPeer peer(input, &connection, this, onConnect);
    peer.run();
    return 0;
}


/*std::string HttpOutput::getStreamTitle()
{
    std::string streamTitle;
    std::string songTitle = input->getSongTitle();
    std::string songArtist = input->getSongArtist();

    if ((songTitle == "") && (songArtist == ""))
        return input->getStreamDescr();
    
    if ((songTitle != "") && (songArtist != ""))
        streamTitle = songArtist + " - " + songTitle;
    else
        if (songTitle != "")
            streamTitle = songTitle;
        else
            if (songArtist != "")
                streamTitle = songArtist;
    
    return streamTitle;
}*/


std::set<MediaType> HttpOutput::getAcceptedMediaTypes()
{
    std::set<MediaType> mtypes;
    mtypes.insert(MediaType("*", "*"));
    return mtypes;
}



Unit* HttpOutputFactory::createUnit(const std::string &name, HTable conf)
{
    if (name != "httpoutput")
        throw MasException("Invalid httpoutput unit type'" + name + "'");
    if (conf)
        return new HttpOutput(conf);
    else
        return new HttpOutput();
}


extern "C" int EXPORT_REGFUNC register_httpoutput_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addUnitFactory(std::string("httpoutput"), 
                new HttpOutputFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

