#include "mp3fileformat.h"


static void check_for_id3v2(std::ifstream& ifs, long& start)
{
    char buf[10];
    int res;
    
    ifs.seekg(0L, std::ios::beg);
    ifs.read(buf, sizeof(buf));
    if (! ifs.good())
        throw MasException("Error reading mp3 file info");
    if (! strncmp("ID3", buf, 3))
    {
        res = buf[3+6];
        res |= buf[2+6] << 7;
        res |= buf[1+6] << 14;
        res |= buf[0+6] << 21;
        if (res > 0)
            start = res;
    }
}


static long getFileSize(std::ifstream& ifs)
{
    ifs.seekg(0, std::ios::end);
    return ifs.tellg();
}


static void strip_buf(char *buf)
{
    int len;
        
    len = strlen(buf);
    while ((len > 0) && (buf[len - 1] == ' '))
        len--;
    buf[len] = 0;
}


static void check_for_id3v1(std::ifstream &ifs, long &size, MediaInfo &info)
{
    char buf[3], bbuf[31];
    
    ifs.seekg(-128, std::ios::end);
    ifs.read(buf, sizeof(buf));
    if (! strncmp("TAG", buf, 3))
    {
        size = size - 128;
        
        memset(bbuf, 0, sizeof(bbuf));
        ifs.read(bbuf, 30);
        strip_buf(bbuf);
        info.setProperty("title", bbuf);
        
        memset(bbuf, 0, sizeof(bbuf));
        ifs.read(bbuf, 30);
        strip_buf(bbuf);
        info.setProperty("artist", bbuf);
    }
}


Mp3FileFormat::Mp3FileFormat(const std::string &fname): FileFormat(fname)
{
    start = 0L;
    size = getFileSize(stream);
    if (size < 200)
        throw MasException("Error reading mp3 song info");
    check_for_id3v2(stream, start);
    check_for_id3v1(stream, size, minfo);
    size -= start;
}


#define BLOCK_SIZE 4096


void Mp3FileFormat::packetize(Interruptable *intr, PacketizerReceiver *receiver)
{
    if (! receiver)
        throw MasException("Invalid mp3 file packets recevier");
    
    Packetizer *packetizer = Server::getServer()->createPacketizer("mp3", receiver);
    if (! packetizer)
        throw MasException("mp3 packetizer not found");
    
    unsigned char buf[BLOCK_SIZE];
    long long int pos;
    std::streamsize toRead, readed;

    stream.seekg(start, std::ios::beg);
    if (! stream.good())
    {
        delete packetizer;
        throw MasException("Error positioning in file '" + fullName + "'");
    }

    pos = 0;
    while ((pos < size) && ((! intr) || (! intr->isShuttingDown())))
    {
        toRead = BLOCK_SIZE;
        if (pos + toRead > size)
            toRead = size - pos;
        stream.read((char*)buf, toRead);
        readed = stream.gcount();
        if (stream.good() && (readed > 0))
            packetizer->addData(buf, readed);
        if ((! stream.good()) || (readed != toRead))
        {
            delete packetizer;
            throw MasException("Error reading file '" + fullName + "'");
        }
        pos += readed;
    }

    delete packetizer;
}


bool Mp3FileFormat::hasMediaInfo()
{
    return (! minfo.isEmpty());
}


MediaInfo Mp3FileFormat::getMediaInfo()
{
    return minfo;
}


bool Mp3FileFormatFactory::isOurFileByExt(const std::string &fileExtension)
{
    return fileExtension == "mp3";
}

bool Mp3FileFormatFactory::isOurFileByContent(std::ifstream &stream,
        const std::string &fullname, const std::string &ext)
{
    return isOurFileByExt(ext);
}

FileFormat* Mp3FileFormatFactory::createFileFormat(const std::string &fullname)
{
    return new Mp3FileFormat(fullname);
}


extern "C" int EXPORT_REGFUNC register_mp3fileformat_plugins()
{
    int res = 0;
    try
    {
        FileFormatRegistry *registry = (FileFormatRegistry*)
            Server::getServer()->findPlugin(std::string("fileformats"));
        if (! registry)
            throw MasException("Can't find file format registry plugin");
        registry->registerFactory(new Mp3FileFormatFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

