#include "rtpinput.h"


#define MAX_PACKET_SIZE 40960


RtpInput::RtpInput(): OutputUnit()
{
    rtpSock = NULL;
    annSock = NULL;
    frameBuf = NULL;
    parser = NULL;
    registry = (RtpPacketParserRegistry*)
        Server::getServer()->findPlugin("rtpPacketParserRegistry");
    if (! registry)
        throw MasException("rtpPacketParserRegistry plugin not found");
    payload = -1;
}


RtpInput::RtpInput(HTable config): OutputUnit(config)
{
    HTable table;

    if (! config)
        throw MasException("RTP input config missing");

    rtpSock = NULL;
    annSock = NULL;
    
    table = table_get_table(config, "rtpsocket", NULL, NULL);
    if (! table)
        throw MasException("RTP socket not specified");
    rtpSock = createReadSocket(table_get_strs(table, "socktype", "", NULL), table);

    table = table_get_table(config, "annsocket", NULL, NULL);
    if (table)
    {
        try
        {
            annSock = createReadSocket(table_get_strs(table, "socktype", "", NULL),
                    table);
        }
        catch (MasException& e)
        {
            delete rtpSock;
            throw e;
        }
        catch (...)
        {
            delete rtpSock;
            throw MasException("Error creating announcer socket");
        }
    }

    parser = NULL;
    registry = (RtpPacketParserRegistry*)
        Server::getServer()->findPlugin("rtpPacketParserRegistry");
    if (! registry)
    {
        if (rtpSock) delete rtpSock;
        if (annSock) delete annSock;
        throw MasException("rtpPacketParserRegistry plugin not found");
    }
    
    maxFrameSize = MAX_PACKET_SIZE;
    frameBuf = new unsigned char[maxFrameSize];
}


RtpInput::~RtpInput()
{
    if (parser) delete parser;
    if (rtpSock) delete rtpSock;
    if (annSock) delete annSock;
    if (frameBuf) delete[] frameBuf;
}


std::set<MediaType> RtpInput::getAcceptedMediaTypes()
{
    std::set<MediaType> s;
    s.insert(MediaType("*", "*"));
    return s;
}


void RtpInput::processHeader(const std::string& name, const std::string& value)
{
//std::cout << "'" << name << "': '" << value << "'" << std::endl;
    if (name == "x-audiocast-streamtitle")
        setStreamName(value);
    else if (name == "x-audiocast-streamurl")
        setStreamUrl(value);
}


void RtpInput::parseHeader(const std::string& line)
{
    int i = 0;
    int len = line.length();
    
//std::cout << "header '" << line << "'" << std::endl;
    while (i < len)
    {
        if (line[i] == ':')
        {
            int nameEnd = i;
            i++;
            while ((i < len) && ((line[i] == ' ') || (line[i] == '\t')))
                i++;
            std::string name = toLowerCase(line.substr(0, nameEnd));
            std::string value = line.substr(i);
            processHeader(name, value);
            break;
        }
        i++;
    }
}


void RtpInput::readAnnPacket()
{
    int len = annSock->read(frameBuf, MAX_PACKET_SIZE);
    if (len <= 0)
        return;
    
    int strStart = 0;
    int i = 0;
    for (i = 0; i < len; i++)
    {
        if (frameBuf[i] == '\n')
        {
            if (i - strStart > 1)
                parseHeader(std::string((char*)frameBuf + strStart, i - strStart));
            strStart = i + 1;
        }
    }
    if (i - strStart > 1)
        parseHeader(std::string((char*)frameBuf + strStart, i - strStart));
}


void RtpInput::readRtpPacket()
{
    int len = rtpSock->read(frameBuf, MAX_PACKET_SIZE);
    if (len < 17)
        return;
    int pl = frameBuf[1] & 0x7F;
 
    if ((! parser) || (pl != payload))
    {
        if (parser)
        {
            parser->flush();
            delete parser;
            parser = NULL;
        }
        payload = pl;
        parser = registry->createParser(payload, this);
    }
    
    if (parser)
        parser->parsePacket(frameBuf + 16, len - 16);
}




void RtpInput::run()
{
    Selector selector;
    struct timeval tv;
    
    while (! isShuttingDown())
    {
//std::cout << "RtpInput::run 1" << std::endl;
        try
        {
            selector.clear();
            selector.addReadSocket(*rtpSock);
            if (annSock)
                selector.addReadSocket(*annSock);
            tv.tv_sec = 1;
            tv.tv_usec = 0;
//std::cout << "RtpInput::run 2" << std::endl;
            if (selector.waitForEvent(tv) > 0)
            {
//std::cout << "RtpInput::run 3" << std::endl;
                if (selector.canRead(*rtpSock))
                    readRtpPacket();
                if (annSock && selector.canRead(*annSock))
                {
//std::cout << "RtpInput::run 3.5" << std::endl;
                    readAnnPacket();
                }
//std::cout << "RtpInput::run 4" << std::endl;
            }
        }
        catch (MasException& e)
        {
            masLog.error(e.getMessage());
        }
        catch (...)
        {
            masLog.error("Error playing file");
        }
    }
//std::cout << "RtpInput::run 5" << std::endl;
}


/*void RtpInput::addPacket(unsigned char *data, int length, double playTime)
{
    buffer->addData(data, length, playTime);
}*/

void RtpInput::addData(unsigned char *data, int length, double playTime, 
        const MediaType &mediaType)
{
    struct timeval tv;

    if (output)
        output->addData(data, length, playTime, mediaType);

    gettimeofday(&tv);
    playingTime = playingTime + playTime / 100;
    unsigned long long int realPlayTime = (tv.tv_sec * 10000 + tv.tv_usec / 100) -
        startTime;
    long long int pause = (long long int)(playingTime - realPlayTime) * 100;
    pause -= 1000000;
    if (pause > 0)
        usleep(pause);
}




Unit* RtpInputUnitFactory::createUnit(const std::string &name, HTable conf)
{
    if (name != "rtpinput")
        throw MasException("Invalid rtp input unit type'" + name + "'");
    if (conf)
        return new RtpInput(conf);
    else
        return new RtpInput();
}


extern "C" int EXPORT_REGFUNC register_rtpinput_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addUnitFactory(std::string("rtpinput"), 
                new RtpInputUnitFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

