#include "rtpoutput.h"
#include "rtp.h"



Announcer::Announcer(const MediaInfo &minfo, WriteSocket *s)
{
    mediaInfo = minfo;
    socket = s;
}


Announcer::~Announcer()
{
}


void Announcer::run()
{
    std::string streamTitle;
    std::string songTitle;
    std::string songArtist;
    std::string descr /*= input->getStreamDescr() */;
    std::string url /*= input->getStreamUrl() */;
    std::string buf;
 
    while (! isShuttingDown())
    {
        mediaLock.lock();
        MediaInfo minfo(mediaInfo);
        mediaLock.unlock();
        streamTitle = "";

        songTitle = minfo.getProperty("title");
        songArtist = minfo.getProperty("artist");
        descr = minfo.getProperty("name");

        if ((songTitle != "") && (songArtist != ""))
            streamTitle = songArtist + " - " + songTitle;
        else
            if (songTitle != "")
                streamTitle = songTitle;
            else
                if (songArtist != "")
                    streamTitle = songArtist;
        
        if (streamTitle == "")
            streamTitle = descr;
        else
            if (descr != "")
                streamTitle = streamTitle + " (" + descr + ")";

        if (streamTitle != "")
            buf = "x-audiocast-streamtitle: " + streamTitle + "\n";
        else 
            buf = "";
        if (url != "")
            buf = buf + "x-audiocast-streamurl: " + url + "\n";
        
        if (buf != "")
            socket->write((unsigned char*)buf.c_str(), buf.length());

        if (! isShuttingDown())
        {
            mediaLock.lock();
            if (minfo == mediaInfo)
            {
                mediaLock.unlock();
                waitForShutdown(20);
            }
            else
                mediaLock.unlock();
        }
    }
}

void Announcer::setMediaInfo(const MediaInfo &minfo)
{
    Locker locker(mediaLock);
    mediaInfo = minfo;
    stopCond.signal();
}



RtpOutput::RtpOutput(): SingleInputUnit()
{
    announcer = NULL;
    annSock = NULL;
    rtpSock = NULL;
}


RtpOutput::RtpOutput(HTable params): SingleInputUnit(params)
{
    HTable table;
    
    table = table_get_table(params, "rtpsocket", NULL, NULL);
    if (! table)
        throw MasException("RTP socket not specified");
    rtpSock = createWriteSocket(table_get_strs(table, "socktype", "", NULL), table);

    table = table_get_table(params, "annsocket", NULL, NULL);
    announcer = NULL;
    annSock = NULL;
    if (table)
    {
        try
        {
            annSock = createWriteSocket(table_get_strs(table, "socktype", "", NULL),
                    table);
            MediaInfo mediaInfo;
            patchMediaInfo(mediaInfo);
            announcer = new Announcer(mediaInfo, annSock);
        }
        catch (MasException& e)
        {
            if (annSock)
                delete annSock;
            delete rtpSock;
            throw e;
        }
    }
}


RtpOutput::~RtpOutput()
{
    if (announcer)
        delete announcer;
    if (annSock)
        delete annSock;
    if (rtpSock)
        delete rtpSock;
}


#define PACKET_BUFFER_SIZE 2048


void RtpOutput::run()
{
    if (! rtpSock)
        return;
    
    if (! input)
        return;
    
    unsigned char *buf;
    RtpSender rtpSender(*rtpSock);
    int readed, bufSize = 1024;
    unsigned long long int pos = 0;
    double playTime;
    MediaType *mediaType=NULL, *mtype;
    MediaInfo *mediaInfo, minfo, patchedMinfo;
    
    patchMediaInfo(patchedMinfo);
    
    if (announcer)
        announcer->start();
    
    buf = (unsigned char*)malloc(bufSize);
    if (! buf)
        throw MasException("Error allocating http peer buffer");
    
    announcer->setMediaInfo(patchedMinfo);
    
    while (! isShuttingDown())
    {
        input->beginGet();
        input->getPacketInfo(mtype, mediaInfo, readed, playTime, pos);
        if (readed > 0)
        {
            if (! (*mediaInfo == minfo))
            {
                minfo = *mediaInfo;
                patchedMinfo = minfo;
                patchMediaInfo(patchedMinfo);
                if (announcer)
                    announcer->setMediaInfo(patchedMinfo);
            }
            if ((! mediaType) || (! mtype->isEqual(*mediaType)))
            {
                if (mediaType) delete mediaType;
                mediaType = mtype->clone();
            }
            if (readed > bufSize)
            {
                bufSize = readed;
                unsigned char *d = (unsigned char*)realloc(buf, bufSize);
                if (! d)
                {
                    input->endGet();
                    free(buf);
                    if (mediaType) delete mediaType;
                    throw MasException("Error growing http peer buffer");
                }
            }
            input->getData(pos, buf);
        }
        input->endGet();
        if (readed > 0)
            rtpSender.sendPacket(buf, readed, playTime, *mediaType);
    }

    rtpSender.flush();

    if (announcer)
        announcer->stop();

    if (mediaType) delete mediaType;
}


std::set<MediaType> RtpOutput::getAcceptedMediaTypes()
{
    std::set<MediaType> mtypes;
    mtypes.insert(MediaType("*", "*"));
    return mtypes;
}



Unit* RtpOutputFactory::createUnit(const std::string &name, HTable conf)
{
    if (name != "rtpoutput")
        throw MasException("Invalid rtpoutput unit type'" + name + "'");
    if (conf)
        return new RtpOutput(conf);
    else
        return new RtpOutput();
}


extern "C" int EXPORT_REGFUNC register_rtpoutput_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addUnitFactory(std::string("rtpoutput"), 
                new RtpOutputFactory());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

