#include <stdlib.h>
#include <sys/time.h>
#include "rtp.h"


RtpSender::RtpSender(WriteSocket& sock, int packetSize): wsock(sock)
{
    ssrc = rndGen.genInt32() & 0xFFFF;
    headerGap = 16;
    maxPacketSize = packetSize - headerGap;
    buffer = new unsigned char[maxPacketSize + headerGap];
    sequenceNo = 0;
    rtpTimestamp = rndGen.genInt31();
    //resetBaseTime();
    rtpTimestamp = 0;
    producer = NULL;
    mediaType = NULL;
    registry = (RtpPacketProducerRegistry*)
        Server::getServer()->findPlugin("rtpPacketProducerRegistry");
    if (! registry)
        throw MasException("rtpPacketProducerRegistry plugin not found");
}


RtpSender::~RtpSender()
{    
    flush();
    if (producer)
        delete producer;
    if (mediaType)
        delete mediaType;
    delete[] buffer;
}


/*void RtpSender::resetBaseTime()
{
#ifdef WIN32
    int Idunno;
#else
    struct timezone Idunno;
#endif
    struct timeval curTime;
    gettimeofday(&curTime, &Idunno);
    baseTime = curTime.tv_sec * 1000000 + curTime.tv_usec;
}*/


void RtpSender::sendPacket(unsigned char *data, int length, double playTime, 
        const MediaType &mtype)
{
    if (length > maxPacketSize)
        throw MasException("Packet is too large"); /// TODO: fragment it
   
    if (mediaType)
    {
        if (! mtype.isEqual(*mediaType))
        {
            if (producer)
            {
                producer->flush();
                delete producer;
                producer = NULL;
            }
            delete mediaType;
            mediaType = NULL;
        }
    }
    if (! mediaType)
    {
        mediaType = mtype.clone();
        if (producer)
        {
            producer->flush();
            delete producer;
            producer = NULL;
        }
    }
    
    if (! producer)
    {
        producer = registry->createProducer(this, *mediaType);
        if (! producer)
            throw MasException("Can't find rtp producer for " + mediaType->getType() 
                    + "/" + mediaType->getSubtype() + " media type");
    }

    producer->makePackets(data, length, playTime, *mediaType);
}


void RtpSender::sendData(unsigned char *data, int length, double playTime, 
        int payload)
{
    if (length > maxPacketSize)
        throw MasException("Packet is too large"); /// TODO: fragment it
   
    memcpy(buffer + headerGap, data, length);
    makePacket(payload, playTime);
    int len, sended;
    len = headerGap + length;
    sended = wsock.write(buffer, len);
}


static void uintToBuf(unsigned char *buf, unsigned int num)
{
    for (int i = 0; i < 4; i++)
    {
        buf[i] = (num & 0xFF000000) >> 24;
        num = num << 8;
    }
}


void RtpSender::makePacket(int payload, double playTime)
{
    buffer[0] = 0x80;   // 10000000
    buffer[1] = payload & 0x7F;
    buffer[2] = (sequenceNo >> 8) & 0xff;
    buffer[3] = sequenceNo & 0xff;

    sequenceNo++;

    rtpTimestamp += (int)((playTime / 100.0) * 9.0);
    uintToBuf(buffer + 4, rtpTimestamp);
    uintToBuf(buffer + 8, ssrc);
    uintToBuf(buffer + 12, 0);
}


void RtpSender::flush()
{
    if (producer)
        producer->flush();
}


RtpPacketProducerRegistry::~RtpPacketProducerRegistry()
{
    for (std::list<RtpPacketProducerFactory*>::iterator i = factories.begin();
            i != factories.end(); i++)
    {
        delete *i;
    }
}


void RtpPacketProducerRegistry::registerFactory(RtpPacketProducerFactory *factory)
{
    factories.push_back(factory);
}


RtpPacketProducer* RtpPacketProducerRegistry::createProducer(RtpSender *sender,
        const MediaType &mediaType)
{
    RtpPacketProducerFactory* f;
    for (std::list<RtpPacketProducerFactory*>::iterator i = factories.begin();
            i != factories.end(); i++)
    {
        f = (*i);
        if (f)
            if (f->isSupportedMediaType(mediaType))
                return f->createPacketProducer(sender, mediaType);
    }
    throw MasException("No rtp producer for " + mediaType.getType() + "/" +
            mediaType.getSubtype() + " media type");
}


RtpPacketParserRegistry::~RtpPacketParserRegistry()
{
    for (std::map<int, RtpPacketParserFactory*>::iterator i = factories.begin();
            i != factories.end(); i++)
        delete (*i).second;
}


void RtpPacketParserRegistry::registerFactory(int payload, RtpPacketParserFactory *factory)
{
    if ((! factory) || (factories.count(payload) > 0))
        throw MasException("Attempt of registering invalit mp3 parser factory");
    factories[payload] = factory;
}


RtpPacketParser* RtpPacketParserRegistry::createParser(int payload, 
        PacketizerReceiver *recv)
{
    if (factories.count(payload) < 1)
        throw MasException("No rtp parser for " + numToStr(payload) + " payload type");
    RtpPacketParserFactory* f = factories[payload];
    return f->createParser(payload, recv);
}


extern "C" int EXPORT_REGFUNC register_rtp_plugins()
{
    int res = 0;
    try
    {
        Server::getServer()->addPlugin(std::string("rtpPacketProducerRegistry"), 
                new RtpPacketProducerRegistry());
        Server::getServer()->addPlugin(std::string("rtpPacketParserRegistry"), 
                new RtpPacketParserRegistry());
    }
    catch (MasException &e)
    {
        masLog.error(e.getMessage());
        res = 1;
    }
    catch (...)
    {
        res = 1;
    }
    return res;
}

